from __future__ import annotations

import os
import posixpath
from collections.abc import Callable
from datetime import datetime, timezone
from enum import Enum
from typing import (
    Any,
    TypeAlias,
    TypeVar,
    cast,
)
from urllib.parse import ParseResult as URLParseResult
from urllib.parse import urljoin, urlparse, urlunparse

from pystac.errors import RequiredPropertyMissing

#: HREF string or path-like object.
HREF: TypeAlias = str | os.PathLike[str]


def make_posix_style(href: HREF) -> str:
    """Converts double backslashes and single backslashes to single forward
    slashes for converting Windows paths to Posix style.

    Args:
        href (str | os.PathLike) : The href string or path-like object.

    Returns:
        str : The converted href in string form.
    """
    _href = str(os.fspath(href))
    return _href.replace("\\\\", "/").replace("\\", "/")


def safe_urlparse(href: str) -> URLParseResult:
    """Wrapper around :func:`urllib.parse.urlparse` that returns consistent results for
    both Windows and UNIX file paths.

    For Windows paths, this function will include the drive prefix (e.g. ``"D:\\"``) as
    part of the ``path`` of the :class:`urllib.parse.ParseResult` rather than as the
    ``scheme`` for consistency with handling of UNIX/LINUX file paths.

    Args:
        href (str) : The HREF to parse. May be a local file path or URL.

    Returns:
        urllib.parse.ParseResult : The named tuple representing the parsed HREF.
    """
    parsed = urlparse(href)
    if parsed.scheme != "" and (
        href.lower().startswith(f"{parsed.scheme}:\\")
        or (
            href.lower().startswith(f"{parsed.scheme}:/")
            and not href.lower().startswith(f"{parsed.scheme}://")
        )
    ):
        return URLParseResult(
            scheme="",
            netloc="",
            path="{}:{}".format(
                # We use this more complicated formulation because parsed.scheme
                # converts to lower-case
                href[: len(parsed.scheme)],
                parsed.path,
            ),
            params=parsed.params,
            query=parsed.query,
            fragment=parsed.fragment,
        )

    # Windows drives sometimes get parsed as the netloc and sometimes
    # as part of the parsed.path.
    if parsed.scheme == "file" and os.name == "nt":
        if parsed.netloc:
            path = f"{parsed.netloc}{parsed.path}"
        elif parsed.path.startswith("/") and ":" in parsed.path:
            path = parsed.path[1:]
        else:
            path = parsed.path

        return URLParseResult(
            scheme=parsed.scheme,
            netloc="",
            path=path,
            params=parsed.params,
            query=parsed.query,
            fragment=parsed.fragment,
        )
    else:
        return parsed


class StringEnum(str, Enum):
    """Base :class:`enum.Enum` class for string enums that will serialize as the string
    value."""

    def __repr__(self) -> str:
        return repr(self.value)

    def __str__(self) -> str:
        return cast(str, self.value)


class JoinType(StringEnum):
    """DEPRECATED.

    .. deprecated:: 1.8.0
        No longer used internally by pystac.

    Allowed join types for :func:`~pystac.utils.join_path_or_url`.
    """

    @staticmethod
    def from_parsed_uri(parsed_uri: URLParseResult) -> JoinType:
        """DEPRECATED.

        .. deprecated:: 1.8.0
            No longer used internally by pystac.

        Determines the appropriate join type based on the scheme of the parsed
        result.

        Args:
            parsed_uri (urllib.parse.ParseResult) : A named tuple representing the
                parsed URI.

        Returns:
            JoinType : The join type for the URI.
        """
        import warnings

        warnings.warn(
            message=(
                "from_parsed_uri is deprecated and will be removed in pystac "
                "version 2.0.0. It is no longer used internally by pystac."
            ),
            category=DeprecationWarning,
        )
        if parsed_uri.scheme == "":
            return JoinType.PATH
        else:
            return JoinType.URL

    PATH = "path"
    URL = "url"


def join_path_or_url(join_type: JoinType, *args: str) -> str:
    """DEPRECATED.

    .. deprecated:: 1.8.0
        No longer used internally by pystac.

    Functions similarly to :func:`os.path.join`, but can be used to join either a
    local file path or a URL.

    Args:
        join_type (JoinType) : One of ``JoinType.PATH`` or ``JoinType.URL``. If
            ``JoinType.PATH``, then :func:`os.path.join` is used for the join.
            If ``JoinType.URL``, then ``posixpath.join`` is used.
        *args (str): Additional positional string arguments to be joined.

    Returns:
        str : The joined path
    """
    import warnings

    warnings.warn(
        message=(
            "join_path_or_url is deprecated and will be removed in pystac "
            "version 2.0.0. It is no longer used internally by pystac."
        ),
        category=DeprecationWarning,
    )
    if join_type == JoinType.PATH:
        return os.path.join(*args)
    else:
        return posixpath.join(*args)


def _make_relative_href_url(
    parsed_source: URLParseResult,
    parsed_start: URLParseResult,
    start_is_dir: bool = False,
) -> str:
    # If the start path is not a directory, get the parent directory
    start_dir = (
        parsed_start.path if start_is_dir else os.path.dirname(parsed_start.path)
    )

    # Strip the leading slashes from both paths
    start_dir = start_dir.lstrip("/")
    source_path = parsed_source.path.lstrip("/")

    # Get the relative path
    rel_url = posixpath.relpath(source_path, start_dir)

    # Ensure we retain a trailing slash from the original source path
    if parsed_source.path.endswith("/"):
        rel_url += "/"

    # Prepend the "./", if necessary
    if rel_url != "./" and not rel_url.startswith("../"):
        rel_url = "./" + rel_url

    return rel_url


def _make_relative_href_path(
    parsed_source: URLParseResult,
    parsed_start: URLParseResult,
    start_is_dir: bool = False,
) -> str:
    # If the start path is not a directory, get the parent directory
    start_dir = (
        parsed_start.path if start_is_dir else os.path.dirname(parsed_start.path)
    )

    # Strip the leading slashes from both paths
    start_dir = start_dir.lstrip("/")
    source_path = parsed_source.path.lstrip("/")

    # posixpath doesn't play well with windows drive letters, so we have to use
    # the os-specific path library for the relpath function. This means we can
    # only handle windows paths on windows machines.
    relpath = make_posix_style(os.path.relpath(source_path, start_dir))

    # Ensure we retain a trailing slash from the original source path
    if parsed_source.path.endswith("/"):
        relpath += "/"

    if relpath != "./" and not relpath.startswith("../"):
        relpath = "./" + relpath

    return relpath


def make_relative_href(
    source_href: str, start_href: str, start_is_dir: bool = False
) -> str:
    """Returns a new string that represents the ``source_href`` as a path relative to
    ``start_href``. If ``source_href`` and ``start_href`` do not share a common parent,
    then ``source_href`` is returned unchanged.

    May be used on either local file paths or URLs.

    Args:
        source_href : The HREF to make relative.
        start_href : The HREF that the resulting HREF will be relative to.
        start_is_dir : If ``True``, ``start_href`` is treated as a directory.
            Otherwise, ``start_href`` is considered to be a path to a file. Defaults to
            ``False``.

    Returns:
        str: The relative HREF.
    """
    source_href = make_posix_style(source_href)
    start_href = make_posix_style(start_href)

    parsed_source = safe_urlparse(source_href)
    parsed_start = safe_urlparse(start_href)
    if not (
        parsed_source.scheme == parsed_start.scheme
        and parsed_source.netloc == parsed_start.netloc
    ):
        return source_href

    if parsed_start.scheme in ["", "file"]:
        return _make_relative_href_path(parsed_source, parsed_start, start_is_dir)
    else:
        return _make_relative_href_url(parsed_source, parsed_start, start_is_dir)


def _make_absolute_href_url(
    parsed_source: URLParseResult,
    parsed_start: URLParseResult,
    start_is_dir: bool = False,
) -> str:
    # If the source is already absolute, just return it
    # We also treat /vsi paths, from GDAL, as absolute
    if parsed_source.scheme != "" or parsed_source.path.startswith("/vsi"):
        return urlunparse(parsed_source)

    # If the start path is not a directory, get the parent directory
    if start_is_dir:
        start_dir = parsed_start.path
    else:
        # Ensure the directory has a trailing slash so urljoin works properly
        start_dir = parsed_start.path.rsplit("/", 1)[0] + "/"

    # Join the start directory to the relative path and find the absolute path
    abs_path = urljoin(start_dir, parsed_source.path)
    abs_path = abs_path.replace("\\", "/")

    return urlunparse(
        (
            parsed_start.scheme,
            parsed_start.netloc,
            abs_path,
            parsed_source.params,
            parsed_source.query,
            parsed_source.fragment,
        )
    )


def _make_absolute_href_path(
    parsed_source: URLParseResult,
    parsed_start: URLParseResult,
    start_is_dir: bool = False,
) -> str:
    # If the source is already absolute, just return it
    if os.path.isabs(parsed_source.path):
        return urlunparse(parsed_source)

    # If the start path is not a directory, get the parent directory
    start_dir = (
        parsed_start.path if start_is_dir else os.path.dirname(parsed_start.path)
    )

    # Join the start directory to the relative path and find the absolute path
    abs_path = make_posix_style(
        os.path.abspath(os.path.join(start_dir, parsed_source.path))
    )

    # Account for the normalization of abspath for
    # things like /vsitar// prefixes by replacing the
    # original start_dir text when abspath modifies the start_dir.
    if not start_dir == make_posix_style(os.path.abspath(start_dir)):
        abs_path = abs_path.replace(
            make_posix_style(os.path.abspath(start_dir)), start_dir
        )

    if parsed_source.scheme or parsed_start.scheme:
        abs_path = f"file://{abs_path}"

    return abs_path


def make_absolute_href(
    source_href: str, start_href: str | None = None, start_is_dir: bool = False
) -> str:
    """Returns a new string that represents ``source_href`` as an absolute path. If
    ``source_href`` is already absolute it is returned unchanged. If ``source_href``
    is relative, the absolute HREF is constructed by joining ``source_href`` to
    ``start_href``.

    May be used on either local file paths or URLs.

    Args:
        source_href : The HREF to make absolute.
        start_href : The HREF that will be used as the basis for resolving relative
            paths, if ``source_href`` is a relative path. Defaults to the current
            working directory.
        start_is_dir : If ``True``, ``start_href`` is treated as a directory.
            Otherwise, ``start_href`` is considered to be a path to a file. Defaults to
            ``False``.

    Returns:
        str: The absolute HREF.
    """
    if start_href is None:
        start_href = os.getcwd()
        start_is_dir = True

    source_href = make_posix_style(source_href)
    start_href = make_posix_style(start_href)

    parsed_start = safe_urlparse(start_href)
    parsed_source = safe_urlparse(source_href)

    if parsed_source.scheme not in ["", "file"] or parsed_start.scheme not in [
        "",
        "file",
    ]:
        return _make_absolute_href_url(parsed_source, parsed_start, start_is_dir)
    else:
        return _make_absolute_href_path(parsed_source, parsed_start, start_is_dir)


def is_absolute_href(href: str, start_href: str | None = None) -> bool:
    """Determines if an HREF is absolute or not.

    May be used on either local file paths or URLs.

    Args:
        href : The HREF to consider.
        start_href : The HREF that will be used as the basis for checking if
          ``source_href`` is a relative path. Defaults to None.

    Returns:
        bool: ``True`` if the given HREF is absolute, ``False`` if it is relative.
    """
    parsed = safe_urlparse(href)
    # We treat /vsi paths, from GDAL, as absolute
    if parsed.scheme not in ["", "file"] or parsed.path.startswith("/vsi"):
        return True
    else:
        parsed_start_scheme = (
            "" if start_href is None else safe_urlparse(start_href).scheme
        )
        return parsed_start_scheme in ["", "file"] and os.path.isabs(parsed.path)


def datetime_to_str(dt: datetime, timespec: str = "auto") -> str:
    """Converts a :class:`datetime.datetime` instance to an ISO8601 string in the
    `RFC 3339, section 5.6
    <https://datatracker.ietf.org/doc/html/rfc3339#section-5.6>`__ format required by
    the :stac-spec:`STAC Spec <master/item-spec/common-metadata.md#date-and-time>`.

    Args:
        dt : The datetime to convert.
        timespec: An optional argument that specifies the number of additional
            terms of the time to include. Valid options are 'auto', 'hours',
            'minutes', 'seconds', 'milliseconds' and 'microseconds'. The default value
            is 'auto'.

    Returns:
        str: The ISO8601 (RFC 3339) formatted string representing the datetime.
    """
    if dt.tzinfo is None:
        dt = dt.replace(tzinfo=timezone.utc)

    timestamp = dt.isoformat(timespec=timespec)
    zulu = "+00:00"
    if timestamp.endswith(zulu):
        timestamp = f"{timestamp[: -len(zulu)]}Z"

    return timestamp


def str_to_datetime(s: str) -> datetime:
    """Converts a string timestamp to a :class:`datetime.datetime` instance using
    :meth:`dateutil.parser.parse` under the hood. The input string may be in any
    format supported by the parser.parser>. This includes many formats including
    ISO 8601 and RFC 3339.

    Args:
        s (str) : The string to convert to :class:`datetime.datetime`.

    Returns:
        str: The :class:`datetime.datetime` represented the by the string.
    """
    import dateutil.parser

    return dateutil.parser.isoparse(s)


def now_in_utc() -> datetime:
    """Returns a datetime value of now with the UTC timezone applied"""
    return datetime.now(timezone.utc)


def now_to_rfc3339_str() -> str:
    """Returns an RFC 3339 string representing now"""
    return datetime_to_str(now_in_utc())


def geometry_to_bbox(geometry: dict[str, Any]) -> list[float]:
    """Extract the bounding box from a geojson geometry

    Args:
        geometry : GeoJSON geometry dict

    Returns:
        list: Bounding box of geojson geometry, formatted according to:
        https://tools.ietf.org/html/rfc7946#section-5
    """
    coords = geometry["coordinates"]

    lats: list[float] = []
    lons: list[float] = []

    def extract_coords(coords: list[list[float] | list[list[Any]]]) -> None:
        for x in coords:
            # This handles points
            if isinstance(x, float):
                assert isinstance(coords[0], float), (
                    f"Type mismatch: {coords[0]} is not a float"
                )
                assert isinstance(coords[1], float), (
                    f"Type mismatch: {coords[1]} is not a float"
                )
                lats.append(coords[0])
                lons.append(coords[1])
                return
            if isinstance(x[0], list):
                extract_coords(x)  # type:ignore
            else:
                lat, lon = x
                lats.append(lat)  # type:ignore
                lons.append(lon)  # type:ignore

    extract_coords(coords)

    lons.sort()
    lats.sort()

    bbox = [lats[0], lons[0], lats[-1], lons[-1]]

    return bbox


T = TypeVar("T")
U = TypeVar("U")


def map_opt(fn: Callable[[T], U], v: T | None) -> U | None:
    """Maps the value of an optional type to another value, returning
    ``None`` if the input option is ``None``.

    Args:
        fn (Callable) : A function that maps the non-optional value of type ``T`` to
            another value. This function will be called on non-``None`` values of
            ``v``.
        v (Optional[T]) : The optional value to map.

    Examples:

        Given an optional value like the following...

        .. code-block:: python

            maybe_item: Optional[pystac.Item] = ...

        ...you could replace...

        .. code-block:: python

            maybe_item_id: Optional[str] = None
            if maybe_item is not None:
                maybe_item_id = maybe_item.id

        ...with:

        .. code-block:: python

            maybe_item_id = map_opt(lambda item: item.id, maybe_item)
    """
    return v if v is None else fn(v)


def get_opt(option: T | None) -> T:
    """Retrieves the value of the ``Optional`` type, raising a :exc:`ValueError` if
    the value is ``None``.

    Use this to get a properly typed value from an optional
    in contexts where you can be certain the value is not ``None``.
    If there is potential for a non-``None`` value, it's best to handle
    the ``None`` case of the optional instead of using this method.

    Args:
        option (Optional[T]) : Some ``Optional`` value

    Returns:
        The value of type T wrapped by the Optional[T]

    Examples:

        .. code-block:: python

            d = {
                "some_key": "some_value"
            }

            # This passes
            val: str = get_opt(d.get("some_key"))

            # This raises a ValueError
            val: str = get_opt(d.get("does_not_exist"))
    """
    if option is None:
        raise ValueError("Cannot get value from None")
    return option


def get_required(option: T | None, obj: str | Any, prop: str) -> T:
    """Retrieves an ``Optional`` value that comes from a required property of some
    object. If the option is ``None``, throws an :exc:`pystac.RequiredPropertyMissing`
    with the given obj and property.

    This method is primarily used internally to retrieve properly typed required
    properties from dictionaries. For an example usage, see the
    :attr:`pystac.extensions.eo.Band.name` source code.

    Args:
        option (Optional[T]) : The ``Optional`` value.
        obj (str, Any) : The object from which the value is being retrieved. This will
            be passed to the :exc:`~pystac.RequiredPropertyMissing` exception if
            ``option`` is ``None``.
        prop (str) : The name of the property being retrieved.

    Returns:
        T : The properly typed, non-``None`` value.
    """
    if option is None:
        raise RequiredPropertyMissing(obj, prop)
    return option


def is_file_path(href: str) -> bool:
    """Checks if an HREF resembles a file path.

    This method checks if the given HREF resembles a file path.
    It checks if the path ends with any kind of file extension
    and if true, assumes it is a file.
    Unlike `os.path.isfile()` it does NOT check the actual file.

    Caution: There are cases for which this method may return wrong results!

    Args:
        href (str) : The HREF to consider.

    Returns:
        bool: ``True`` if the given HREF resembles a file path,
            ``False`` if it does not.
    """
    parsed = urlparse(href)
    return bool(os.path.splitext(parsed.path)[1])


def _is_url(href: str) -> bool:
    """Checks if an HREF is a url rather than a local path"""
    parsed = safe_urlparse(href)
    return parsed.scheme not in ["", "file"]



import sys
from typing import Any

import html5lib
import pytest
from pytest_mock import MockerFixture

import pystac
from pystac.html.jinja_env import get_jinja_env
from tests.utils import TestCases


def parse_html(stac_html: str) -> None:
    html5lib.HTMLParser(strict=True).parse("<!DOCTYPE html>\n" + stac_html)


@pytest.mark.parametrize(
    "path,cls",
    [
        ("data-files/item/sample-item.json", pystac.Item),
        (
            "data-files/catalogs/test-case-1/country-1/area-1-1/collection.json",
            pystac.Collection,
        ),
        ("data-files/catalogs/test-case-1/catalog.json", pystac.Catalog),
        (
            "data-files/item-collection/sample-item-collection.json",
            pystac.ItemCollection,
        ),
    ],
)
def test_from_file_valid_html(path: str, cls: Any) -> None:
    href = TestCases.get_path(path)
    obj = cls.from_file(href)
    parse_html(obj._repr_html_())


@pytest.mark.parametrize(
    "path,cls",
    [
        ("data-files/item/sample-item.json", pystac.Item),
        (
            "data-files/catalogs/test-case-1/country-1/area-1-1/collection.json",
            pystac.Collection,
        ),
        ("data-files/catalogs/test-case-1/catalog.json", pystac.Catalog),
        (
            "data-files/item-collection/sample-item-collection.json",
            pystac.ItemCollection,
        ),
    ],
)
def test_from_file_missing_jinja2(mocker: MockerFixture, path: str, cls: Any) -> None:
    get_jinja_env.cache_clear()

    mocker.patch.dict(sys.modules, {"jinja2": None})
    assert get_jinja_env() is None

    href = TestCases.get_path(path)
    obj = cls.from_file(href)
    assert cls.__name__ in obj._repr_html_()

    get_jinja_env.cache_clear()


def test_nested_objects_valid_html() -> None:
    href = TestCases.get_path("data-files/item/sample-item-asset-properties.json")
    item = pystac.Item.from_file(href)

    link = item.links[0]
    parse_html(link._repr_html_())

    asset = item.assets["thumbnail"]
    parse_html(asset._repr_html_())

    provider = (asset.common_metadata.providers or [])[0]
    parse_html(provider._repr_html_())


def test_nested_objects_missing_jinja2(mocker: MockerFixture) -> None:
    get_jinja_env.cache_clear()

    mocker.patch.dict(sys.modules, {"jinja2": None})
    assert get_jinja_env() is None

    href = TestCases.get_path("data-files/item/sample-item-asset-properties.json")
    item = pystac.Item.from_file(href)

    link = item.links[0]
    assert pystac.Link.__name__ in link._repr_html_()

    asset = item.assets["thumbnail"]
    assert pystac.Asset.__name__ in asset._repr_html_()

    provider = (asset.common_metadata.providers or [])[0]
    assert pystac.Provider.__name__ in provider._repr_html_()

    get_jinja_env.cache_clear()



"""Tests for pystac.extensions.sat."""

from datetime import datetime
from typing import Any

import pytest

import pystac
from pystac import Collection, ExtensionTypeError, Item
from pystac.extensions import sat
from pystac.extensions.sat import OrbitState, SatExtension, SummariesSatExtension
from pystac.summaries import RangeSummary
from pystac.utils import datetime_to_str, str_to_datetime
from tests.utils import TestCases


@pytest.fixture
def item() -> Item:
    """Create basic test items that are only slightly different."""
    asset_id = "an/asset"
    start = datetime(2018, 1, 2)
    item = pystac.Item(
        id=asset_id, geometry=None, bbox=None, datetime=start, properties={}
    )

    SatExtension.add_to(item)
    return item


@pytest.fixture
def sentinel_item() -> Item:
    sentinel_example_uri = TestCases.get_path("data-files/sat/sentinel-1.json")
    return pystac.Item.from_file(sentinel_example_uri)


def test_stac_extensions(item: Item) -> None:
    assert SatExtension.has_extension(item)


def test_item_repr(item: Item) -> None:
    sat_item_ext = SatExtension.ext(item)
    assert f"<ItemSatExtension Item id={item.id}>" == sat_item_ext.__repr__()


def test_asset_repr(sentinel_item: Item) -> None:
    asset = sentinel_item.assets["measurement_iw1_vh"]
    sat_asset_ext = SatExtension.ext(asset)

    assert f"<AssetSatExtension Asset href={asset.href}>" == sat_asset_ext.__repr__()


@pytest.mark.vcr()
def test_no_args_fails(item: Item) -> None:
    SatExtension.ext(item).apply()
    with pytest.raises(pystac.STACValidationError):
        item.validate()


@pytest.mark.vcr()
def test_orbit_state(item: Item) -> None:
    orbit_state = sat.OrbitState.ASCENDING
    SatExtension.ext(item).apply(orbit_state)
    assert orbit_state == SatExtension.ext(item).orbit_state
    assert sat.RELATIVE_ORBIT_PROP not in item.properties
    assert SatExtension.ext(item).relative_orbit is None
    item.validate()


@pytest.mark.vcr()
def test_relative_orbit(item: Item) -> None:
    relative_orbit = 1234
    SatExtension.ext(item).apply(None, relative_orbit)
    assert relative_orbit == SatExtension.ext(item).relative_orbit
    assert sat.ORBIT_STATE_PROP not in item.properties
    assert SatExtension.ext(item).orbit_state is None
    item.validate()


@pytest.mark.vcr()
def test_absolute_orbit(item: Item) -> None:
    absolute_orbit = 1234
    SatExtension.ext(item).apply(absolute_orbit=absolute_orbit)
    assert absolute_orbit == SatExtension.ext(item).absolute_orbit
    assert sat.RELATIVE_ORBIT_PROP not in item.properties
    assert SatExtension.ext(item).relative_orbit is None
    item.validate()


@pytest.mark.vcr()
def test_anx_datetime(item: Item) -> None:
    anx_datetime = str_to_datetime("2020-01-01T00:00:00Z")
    SatExtension.ext(item).apply(anx_datetime=anx_datetime)
    assert anx_datetime == SatExtension.ext(item).anx_datetime
    assert sat.RELATIVE_ORBIT_PROP not in item.properties
    assert SatExtension.ext(item).relative_orbit is None
    item.validate()


@pytest.mark.vcr()
def test_platform_international_designator(item: Item) -> None:
    platform_international_designator = "2018-080A"
    SatExtension.ext(item).apply(
        platform_international_designator=platform_international_designator
    )
    assert (
        platform_international_designator
        == SatExtension.ext(item).platform_international_designator
    )
    assert sat.ORBIT_STATE_PROP not in item.properties
    assert SatExtension.ext(item).orbit_state is None
    item.validate()


@pytest.mark.vcr()
def test_relative_orbit_no_negative(item: Item) -> None:
    negative_relative_orbit = -2
    SatExtension.ext(item).apply(None, negative_relative_orbit)
    with pytest.raises(pystac.STACValidationError):
        item.validate()


@pytest.mark.vcr()
def test_both(item: Item) -> None:
    orbit_state = sat.OrbitState.DESCENDING
    relative_orbit = 4321
    SatExtension.ext(item).apply(orbit_state, relative_orbit)
    assert orbit_state == SatExtension.ext(item).orbit_state
    assert relative_orbit == SatExtension.ext(item).relative_orbit
    item.validate()


@pytest.mark.vcr()
def test_modify(item: Item) -> None:
    SatExtension.ext(item).apply(sat.OrbitState.DESCENDING, 999)

    orbit_state = sat.OrbitState.GEOSTATIONARY
    SatExtension.ext(item).orbit_state = orbit_state
    relative_orbit = 1000
    SatExtension.ext(item).relative_orbit = relative_orbit
    assert orbit_state == SatExtension.ext(item).orbit_state
    assert relative_orbit == SatExtension.ext(item).relative_orbit
    item.validate()


def test_from_dict() -> None:
    orbit_state = sat.OrbitState.GEOSTATIONARY
    relative_orbit = 1001
    d: dict[str, Any] = {
        "type": "Feature",
        "stac_version": "1.0.0-beta.2",
        "id": "an/asset",
        "properties": {
            "sat:orbit_state": orbit_state.value,
            "sat:relative_orbit": relative_orbit,
            "datetime": "2018-01-02T00:00:00Z",
        },
        "geometry": None,
        "links": [],
        "assets": {},
        "stac_extensions": [SatExtension.get_schema_uri()],
    }
    item = pystac.Item.from_dict(d)
    assert orbit_state == SatExtension.ext(item).orbit_state
    assert relative_orbit == SatExtension.ext(item).relative_orbit


def test_to_from_dict(item: Item) -> None:
    orbit_state = sat.OrbitState.GEOSTATIONARY
    relative_orbit = 1002
    SatExtension.ext(item).apply(orbit_state, relative_orbit)
    d = item.to_dict()
    assert orbit_state.value == d["properties"][sat.ORBIT_STATE_PROP]
    assert relative_orbit == d["properties"][sat.RELATIVE_ORBIT_PROP]

    item = pystac.Item.from_dict(d)
    assert orbit_state == SatExtension.ext(item).orbit_state
    assert relative_orbit == SatExtension.ext(item).relative_orbit


@pytest.mark.vcr()
def test_clear_orbit_state(item: Item) -> None:
    SatExtension.ext(item).apply(sat.OrbitState.DESCENDING, 999)

    SatExtension.ext(item).orbit_state = None
    assert SatExtension.ext(item).orbit_state is None
    item.validate()


@pytest.mark.vcr()
def test_clear_relative_orbit(item: Item) -> None:
    SatExtension.ext(item).apply(sat.OrbitState.DESCENDING, 999)

    SatExtension.ext(item).relative_orbit = None
    assert SatExtension.ext(item).relative_orbit is None
    item.validate()


def test_extension_not_implemented(sentinel_item: Item) -> None:
    # Should raise exception if Item does not include extension URI
    sentinel_item.stac_extensions.remove(SatExtension.get_schema_uri())

    with pytest.raises(pystac.ExtensionNotImplemented):
        _ = SatExtension.ext(sentinel_item)

    # Should raise exception if owning Item does not include extension URI
    asset = sentinel_item.assets["measurement_iw1_vh"]

    with pytest.raises(pystac.ExtensionNotImplemented):
        _ = SatExtension.ext(asset)

    # Should succeed if Asset has no owner
    ownerless_asset = pystac.Asset.from_dict(asset.to_dict())
    _ = SatExtension.ext(ownerless_asset)


def test_item_ext_add_to(sentinel_item: Item) -> None:
    sentinel_item.stac_extensions.remove(SatExtension.get_schema_uri())
    assert SatExtension.get_schema_uri() not in sentinel_item.stac_extensions

    _ = SatExtension.ext(sentinel_item, add_if_missing=True)

    assert SatExtension.get_schema_uri() in sentinel_item.stac_extensions


def test_asset_ext_add_to(sentinel_item: Item) -> None:
    sentinel_item.stac_extensions.remove(SatExtension.get_schema_uri())
    assert SatExtension.get_schema_uri() not in sentinel_item.stac_extensions
    asset = sentinel_item.assets["measurement_iw1_vh"]

    _ = SatExtension.ext(asset, add_if_missing=True)

    assert SatExtension.get_schema_uri() in sentinel_item.stac_extensions


def test_should_raise_exception_when_passing_invalid_extension_object() -> None:
    with pytest.raises(
        ExtensionTypeError, match=r"^SatExtension does not apply to type 'object'$"
    ):
        # calling this wrong on purpose so ---v
        SatExtension.ext(object())  # type: ignore


@pytest.fixture
def collection() -> Collection:
    return Collection.from_file(
        TestCases.get_path("data-files/collections/multi-extent.json")
    )


@pytest.fixture
def summaries_ext(collection: Collection) -> SummariesSatExtension:
    return SatExtension.summaries(collection, True)


def test_summaries_platform_international_designation(
    collection: Collection, summaries_ext: SummariesSatExtension
) -> None:
    platform_international_designator_list = ["2018-080A"]

    summaries_ext.platform_international_designator = ["2018-080A"]

    assert (
        summaries_ext.platform_international_designator
        == platform_international_designator_list
    )

    summaries_dict = collection.to_dict()["summaries"]

    assert (
        summaries_dict["sat:platform_international_designator"]
        == platform_international_designator_list
    )


def test_summaries_orbit_state(
    collection: Collection, summaries_ext: SummariesSatExtension
) -> None:
    orbit_state_list = [OrbitState.ASCENDING]

    summaries_ext.orbit_state = orbit_state_list

    assert summaries_ext.orbit_state == orbit_state_list

    summaries_dict = collection.to_dict()["summaries"]

    assert summaries_dict["sat:orbit_state"] == orbit_state_list


def test_summaries_absolute_orbit(
    collection: Collection, summaries_ext: SummariesSatExtension
) -> None:
    absolute_orbit_range = RangeSummary(2000, 3000)

    summaries_ext.absolute_orbit = absolute_orbit_range

    assert summaries_ext.absolute_orbit == absolute_orbit_range

    summaries_dict = collection.to_dict()["summaries"]

    assert summaries_dict["sat:absolute_orbit"] == absolute_orbit_range.to_dict()


def test_summaries_relative_orbit(
    collection: Collection, summaries_ext: SummariesSatExtension
) -> None:
    relative_orbit_range = RangeSummary(50, 100)

    summaries_ext.relative_orbit = relative_orbit_range

    assert summaries_ext.relative_orbit == relative_orbit_range

    summaries_dict = collection.to_dict()["summaries"]

    assert summaries_dict["sat:relative_orbit"] == relative_orbit_range.to_dict()


def test_summaries_anx_datetime(
    collection: Collection, summaries_ext: SummariesSatExtension
) -> None:
    anx_datetime_range = RangeSummary(
        str_to_datetime("2020-01-01T00:00:00.000Z"),
        str_to_datetime("2020-01-02T00:00:00.000Z"),
    )

    summaries_ext.anx_datetime = anx_datetime_range

    assert summaries_ext.anx_datetime == anx_datetime_range

    summaries_dict = collection.to_dict()["summaries"]

    assert summaries_dict["sat:anx_datetime"] == {
        "minimum": datetime_to_str(anx_datetime_range.minimum),
        "maximum": datetime_to_str(anx_datetime_range.maximum),
    }


def test_summaries_adds_uri(collection: Collection) -> None:
    collection.stac_extensions = []
    with pytest.raises(
        pystac.ExtensionNotImplemented,
        match="Extension 'sat' is not implemented",
    ):
        SatExtension.summaries(collection, add_if_missing=False)

    _ = SatExtension.summaries(collection, True)

    assert SatExtension.get_schema_uri() in collection.stac_extensions

    SatExtension.remove_from(collection)
    assert SatExtension.get_schema_uri() not in collection.stac_extensions



"""Implements the :stac-ext:`Item Assets Definition Extension <item-assets>`."""

from __future__ import annotations

from typing import Any, Literal

import pystac
from pystac.errors import DeprecatedWarning
from pystac.extensions.base import ExtensionManagementMixin
from pystac.extensions.hooks import ExtensionHooks
from pystac.item_assets import ItemAssetDefinition
from pystac.serialization.identify import STACJSONDescription, STACVersionID
from pystac.utils import get_required

SCHEMA_URI = "https://stac-extensions.github.io/item-assets/v1.0.0/schema.json"

ITEM_ASSETS_PROP = "item_assets"


class AssetDefinition(ItemAssetDefinition):
    """
    DEPRECATED

    .. deprecated:: 1.12.0
        Use :class:`~pystac.ItemAssetDefinition` instead.
    """

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        import warnings

        warnings.warn(
            (
                "``AssetDefinition`` is deprecated. "
                "Please use ``pystac.ItemAssetDefinition`` instead."
            ),
            DeprecationWarning,
        )
        super().__init__(*args, **kwargs)


class ItemAssetsExtension(ExtensionManagementMixin[pystac.Collection]):
    """
    DEPRECATED

    .. deprecated:: 1.12.0
        Use :attr:`~pystac.Collection.item_assets` instead.
    """

    name: Literal["item_assets"] = "item_assets"
    collection: pystac.Collection

    def __init__(self, collection: pystac.Collection) -> None:
        import warnings

        warnings.warn(
            (
                "The ``item_assets`` extension is deprecated. "
                "``item_assets`` is now a top-level property of ``Collection``."
            ),
            DeprecatedWarning,
        )
        self.collection = collection

    @property
    def item_assets(self) -> dict[str, ItemAssetDefinition]:
        """Gets or sets a dictionary of assets that can be found in member Items. Maps
        the asset key to an :class:`AssetDefinition` instance."""
        result: dict[str, Any] = get_required(
            self.collection.extra_fields.get(ITEM_ASSETS_PROP), self, ITEM_ASSETS_PROP
        )
        return {k: ItemAssetDefinition(v, self.collection) for k, v in result.items()}

    @item_assets.setter
    def item_assets(self, v: dict[str, ItemAssetDefinition]) -> None:
        self.collection.extra_fields[ITEM_ASSETS_PROP] = {
            k: asset_def.properties for k, asset_def in v.items()
        }

    def __repr__(self) -> str:
        return f"<ItemAssetsExtension collection.id = {self.collection.id}>"

    @classmethod
    def get_schema_uri(cls) -> str:
        return SCHEMA_URI

    @classmethod
    def ext(
        cls, obj: pystac.Collection, add_if_missing: bool = False
    ) -> ItemAssetsExtension:
        """Extends the given :class:`~pystac.Collection` with properties from the
        :stac-ext:`Item Assets Extension <item-assets>`.

        Raises:

            pystac.ExtensionTypeError : If an invalid object type is passed.
        """
        if isinstance(obj, pystac.Collection):
            cls.ensure_has_extension(obj, add_if_missing)
            return cls(obj)
        else:
            raise pystac.ExtensionTypeError(cls._ext_error_message(obj))


class ItemAssetsExtensionHooks(ExtensionHooks):
    schema_uri: str = SCHEMA_URI
    prev_extension_ids = {"asset", "item-assets"}
    stac_object_types = {pystac.STACObjectType.COLLECTION}

    def migrate(
        self, obj: dict[str, Any], version: STACVersionID, info: STACJSONDescription
    ) -> None:
        # Handle that the "item-assets" extension had the id of "assets", before
        # collection assets (since removed) took over the ID of "assets"
        if version < "1.0.0-beta.1" and "asset" in info.extensions:
            if "assets" in obj:
                obj["item_assets"] = obj["assets"]
                del obj["assets"]

        super().migrate(obj, version, info)

        # As of STAC spec version 1.1.0 item-assets are part of core
        if obj["stac_version"] >= "1.1.0" and self.schema_uri in obj.get(
            "stac_extensions", []
        ):
            obj["stac_extensions"].remove(self.schema_uri)


ITEM_ASSETS_EXTENSION_HOOKS: ExtensionHooks = ItemAssetsExtensionHooks()



from __future__ import annotations

import json
import os
import posixpath
import tempfile
from collections import defaultdict
from collections.abc import Iterator
from copy import deepcopy
from datetime import datetime, timezone
from pathlib import Path
from typing import Any, cast

import pytest

import pystac
from pystac import (
    HIERARCHICAL_LINKS,
    Asset,
    Catalog,
    CatalogType,
    Collection,
    Item,
    Link,
    MediaType,
)
from pystac.errors import STACError
from pystac.layout import (
    APILayoutStrategy,
    BestPracticesLayoutStrategy,
    HrefLayoutStrategy,
    TemplateLayoutStrategy,
)
from pystac.utils import (
    is_absolute_href,
    make_absolute_href,
    make_posix_style,
    make_relative_href,
)
from tests.utils import (
    ARBITRARY_BBOX,
    ARBITRARY_EXTENT,
    ARBITRARY_GEOM,
    MockStacIO,
    TestCases,
)


class TestCatalogType:
    def test_determine_type_for_absolute_published(self) -> None:
        cat = TestCases.case_1()
        with tempfile.TemporaryDirectory() as tmp_dir:
            cat.normalize_and_save(tmp_dir, catalog_type=CatalogType.ABSOLUTE_PUBLISHED)
            cat_json = pystac.StacIO.default().read_json(
                os.path.join(tmp_dir, "catalog.json")
            )

        catalog_type = CatalogType.determine_type(cat_json)
        assert catalog_type == CatalogType.ABSOLUTE_PUBLISHED

    def test_determine_type_for_relative_published(self) -> None:
        cat = TestCases.case_2()
        with tempfile.TemporaryDirectory() as tmp_dir:
            cat.normalize_and_save(tmp_dir, catalog_type=CatalogType.RELATIVE_PUBLISHED)
            cat_json = pystac.StacIO.default().read_json(
                os.path.join(tmp_dir, "catalog.json")
            )

        catalog_type = CatalogType.determine_type(cat_json)
        assert catalog_type == CatalogType.RELATIVE_PUBLISHED

    def test_determine_type_for_self_contained(self) -> None:
        cat_json = pystac.StacIO.default().read_json(
            TestCases.get_path("data-files/catalogs/test-case-1/catalog.json")
        )
        catalog_type = CatalogType.determine_type(cat_json)
        assert catalog_type == CatalogType.SELF_CONTAINED

    def test_determine_type_for_unknown(self) -> None:
        catalog = Catalog(id="test", description="test desc")
        subcat = Catalog(id="subcat", description="subcat desc")
        catalog.add_child(subcat)
        catalog.normalize_hrefs("http://example.com")
        d = catalog.to_dict(include_self_link=False)

        assert CatalogType.determine_type(d) is None


class TestCatalog:
    def test_create_and_read(self) -> None:
        with tempfile.TemporaryDirectory() as tmp_dir:
            cat_dir = os.path.join(tmp_dir, "catalog")
            catalog = TestCases.case_1()

            catalog.normalize_and_save(
                cat_dir, catalog_type=CatalogType.ABSOLUTE_PUBLISHED
            )

            read_catalog = Catalog.from_file(f"{cat_dir}/catalog.json")

            collections = catalog.get_children()
            assert len(list(collections)) == 2

            items = read_catalog.get_items(recursive=True)

            assert len(list(items)) == 8

    def test_from_dict_preserves_dict(self) -> None:
        catalog_dict = TestCases.case_1().to_dict()
        param_dict = deepcopy(catalog_dict)

        # test that the parameter is preserved
        _ = Catalog.from_dict(param_dict)
        assert param_dict == catalog_dict

        # assert that the parameter is not preserved with
        # non-default parameter
        _ = Catalog.from_dict(param_dict, preserve_dict=False, migrate=False)
        assert param_dict != catalog_dict

    def test_from_file_bad_catalog(self) -> None:
        with pytest.raises(pystac.errors.STACTypeError) as ctx:
            _ = Catalog.from_file(TestCases.get_path(TestCases.bad_catalog_case))
        assert "(id = broken_cat) does not represent a STACObject" in ctx.value.args[0]
        assert "is Catalog" in ctx.value.args[0]

    def test_from_dict_set_root(self) -> None:
        path = TestCases.get_path("data-files/catalogs/test-case-1/catalog.json")
        with open(path) as f:
            cat_dict = json.load(f)
        root_cat = pystac.Catalog(id="test", description="test desc")
        collection = Catalog.from_dict(cat_dict, root=root_cat)
        assert collection.get_root() is root_cat

    @pytest.mark.vcr()
    def test_read_remote(self) -> None:
        catalog_url = (
            "https://raw.githubusercontent.com/stac-extensions/label/main/"
            "examples/multidataset/catalog.json"
        )
        cat = Catalog.from_file(catalog_url)

        zanzibar = cat.get_child("zanzibar-collection")
        assert zanzibar is not None

        assert len(list(zanzibar.get_items())) == 2

    def test_clear_items_removes_from_cache(self) -> None:
        catalog = Catalog(id="test", description="test")
        subcat = Catalog(id="subcat", description="test")
        catalog.add_child(subcat)
        item = Item(
            id="test-item",
            geometry=ARBITRARY_GEOM,
            bbox=ARBITRARY_BBOX,
            datetime=datetime.now(timezone.utc),
            properties={"key": "one"},
        )
        subcat.add_item(item)

        items = list(catalog.get_items(recursive=True))
        assert len(items) == 1
        assert items[0].properties["key"] == "one"

        subcat.clear_items()
        item = Item(
            id="test-item",
            geometry=ARBITRARY_GEOM,
            bbox=ARBITRARY_BBOX,
            datetime=datetime.now(timezone.utc),
            properties={"key": "two"},
        )
        subcat.add_item(item)

        items = list(catalog.get_items(recursive=True))
        assert len(items) == 1
        assert items[0].properties["key"] == "two"

        subcat.remove_item("test-item")
        item = Item(
            id="test-item",
            geometry=ARBITRARY_GEOM,
            bbox=ARBITRARY_BBOX,
            datetime=datetime.now(timezone.utc),
            properties={"key": "three"},
        )
        subcat.add_item(item)

        items = list(catalog.get_items(recursive=True))
        assert len(items) == 1
        assert items[0].properties["key"] == "three"

    def test_clear_children_removes_from_cache(self) -> None:
        catalog = Catalog(id="test", description="test")
        subcat = Catalog(id="subcat", description="test")
        catalog.add_child(subcat)

        children = list(catalog.get_children())
        assert len(children) == 1
        assert children[0].description == "test"

        catalog.clear_children()
        subcat = Catalog(id="subcat", description="test2")
        catalog.add_child(subcat)

        children = list(catalog.get_children())
        assert len(children) == 1
        assert children[0].description == "test2"

        catalog.remove_child("subcat")
        subcat = Catalog(id="subcat", description="test3")
        catalog.add_child(subcat)

        children = list(catalog.get_children())
        assert len(children) == 1
        assert children[0].description == "test3"

    def test_clear_children_sets_parent_and_root_to_None(self) -> None:
        catalog = Catalog(id="test", description="test")
        subcat1 = Catalog(id="subcat", description="test")
        subcat2 = Catalog(id="subcat2", description="test2")
        catalog.add_children([subcat1, subcat2])

        assert subcat1.get_parent() is not None
        assert subcat2.get_parent() is not None
        assert subcat1.get_root() is not None
        assert subcat2.get_root() is not None

        children = list(catalog.get_children())
        assert len(children) == 2

        catalog.clear_children()

        assert subcat1.get_parent() is None
        assert subcat2.get_parent() is None
        assert subcat1.get_root() is None
        assert subcat2.get_root() is None

    def test_add_child_throws_if_item(self) -> None:
        cat = TestCases.case_1()
        item = next(cat.get_items(recursive=True))
        with pytest.raises(pystac.STACError):
            cat.add_child(item)  # type:ignore

    def test_add_child_returns_link(self) -> None:
        parent = Catalog(id="parent", description="test")
        child = Catalog(id="child", description="test")
        link = parent.add_child(child)
        assert isinstance(link, pystac.Link)
        link.extra_fields["foo"] = "bar"
        link2 = parent.get_single_link("child")
        assert link2 is not None
        assert link2.extra_fields["foo"] == "bar"

    def test_add_children_returns_links(self) -> None:
        parent = Catalog(id="parent", description="test")
        child1 = Catalog(id="child", description="test")
        child2 = Catalog(id="child2", description="test")
        links = parent.add_children([child1, child2])
        assert isinstance(links, list)
        assert len(links) == 2
        assert isinstance(links[0], pystac.Link)
        assert isinstance(links[1], pystac.Link)
        links2 = parent.get_links("child")
        assert links[0] in links2
        assert links[1] in links2

    def test_add_child_override_parent(self) -> None:
        parent1 = Catalog(id="parent1", description="test1")
        parent2 = Catalog(id="parent2", description="test2")
        child = Catalog(id="child", description="test3")
        assert child.get_parent() is None

        parent1.add_child(child)
        assert child.get_parent() is parent1

        parent2.add_child(child)
        assert child.get_parent() is parent2

    def test_add_child_set_parent_false(self) -> None:
        parent1 = Catalog(id="parent1", description="test1")
        parent2 = Catalog(id="parent2", description="test2")
        child = Catalog(id="child", description="test3")
        assert child.get_parent() is None

        parent1.add_child(child, set_parent=False)
        assert child.get_parent() is not parent1

        parent1.add_child(child)
        parent2.add_child(child, set_parent=False)
        assert child.get_parent() is parent1

    def test_add_item_override_parent(self) -> None:
        parent1 = Catalog(id="parent1", description="test1")
        parent2 = Catalog(id="parent2", description="test2")
        child = Item(
            id="child",
            geometry=None,
            bbox=None,
            datetime=datetime.now(),
            properties={},
        )
        assert child.get_parent() is None

        parent1.add_item(child)
        assert child.get_parent() is parent1

        parent2.add_item(child)
        assert child.get_parent() is parent2

    def test_add_item_set_parent_false(self) -> None:
        parent1 = Catalog(id="parent1", description="test1")
        parent2 = Catalog(id="parent2", description="test2")
        child = Item(
            id="child",
            geometry=None,
            bbox=None,
            datetime=datetime.now(),
            properties={},
        )
        assert child.get_parent() is None

        parent1.add_item(child, set_parent=False)
        assert child.get_parent() is not parent1

        parent1.add_item(child)
        parent2.add_item(child, set_parent=False)
        assert child.get_parent() is parent1

    def test_add_item_throws_if_child(self) -> None:
        cat = TestCases.case_1()
        child = next(iter(cat.get_children()))
        with pytest.raises(pystac.STACError):
            cat.add_item(child)  # type:ignore

    def test_add_item_returns_link(self) -> None:
        parent = Catalog(id="parent", description="test")
        child = Item(
            id="child",
            geometry=None,
            bbox=None,
            datetime=datetime.now(),
            properties={},
        )
        link = parent.add_item(child)
        assert isinstance(link, pystac.Link)
        link.extra_fields["foo"] = "bar"
        link2 = parent.get_single_link("item")
        assert link2 is not None
        assert link2.extra_fields["foo"] == "bar"

    def test_add_items_returns_links(self) -> None:
        parent = Catalog(id="parent", description="test")
        child1 = Item(
            id="child1",
            geometry=None,
            bbox=None,
            datetime=datetime.now(),
            properties={},
        )
        child2 = Item(
            id="child2",
            geometry=None,
            bbox=None,
            datetime=datetime.now(),
            properties={},
        )
        links = parent.add_items([child1, child2])
        assert isinstance(links, list)
        assert len(links) == 2
        assert isinstance(links[0], pystac.Link)
        assert isinstance(links[1], pystac.Link)
        links2 = parent.get_links("item")
        assert links[0] in links2
        assert links[1] in links2

    def test_get_child_returns_none_if_not_found(self) -> None:
        cat = TestCases.case_1()
        child = cat.get_child("thisshouldnotbeachildid", recursive=True)
        assert child is None

    def test_get_item_is_deprecated_but_still_works(self) -> None:
        cat = TestCases.case_1()
        with pytest.warns(DeprecationWarning):
            item = cat.get_item("area-2-1-imagery", recursive=True)
            assert item is not None

    def test_get_item_returns_none_if_not_found(self) -> None:
        cat = TestCases.case_1()
        with pytest.warns(DeprecationWarning):
            item = cat.get_item("thisshouldnotbeanitemid", recursive=True)
            assert item is None

    def test_get_all_items_is_deprecated_but_still_works(self) -> None:
        cat = TestCases.case_1()
        with pytest.warns(DeprecationWarning):
            all_items = cat.get_all_items()
        items = cat.get_items(recursive=True)
        assert all(a == i for a, i in zip(all_items, items))

    def test_get_items_returns_empty_generator_if_not_found(self) -> None:
        cat = TestCases.case_1()
        items = cat.get_items("thisshouldnotbeanitemid", recursive=True)
        assert next(items, None) is None

    def test_sets_catalog_type(self) -> None:
        cat = TestCases.case_1()

        assert cat.catalog_type == CatalogType.SELF_CONTAINED

    @pytest.mark.parametrize("catalog", TestCases.all_test_catalogs())
    def test_walk_iterates_correctly(self, catalog: Catalog) -> None:
        def test_catalog(cat: Catalog) -> None:
            expected_catalog_iterations = 1
            actual_catalog_iterations = 0
            for root, children, items in cat.walk():
                actual_catalog_iterations += 1
                expected_catalog_iterations += len(list(root.get_children()))

                assert {c.id for c in root.get_children()} == {
                    c.id for c in children
                }, "Children unequal"

                assert {c.id for c in root.get_items()} == {c.id for c in items}, (
                    "Items unequal"
                )

            assert actual_catalog_iterations == expected_catalog_iterations

        test_catalog(catalog)

    @pytest.mark.parametrize("catalog", TestCases.all_test_catalogs())
    def test_clone_generates_correct_links(self, catalog: Catalog) -> None:
        expected_link_types_to_counts: Any = {}
        actual_link_types_to_counts: Any = {}

        for root, _, items in catalog.walk():
            expected_link_types_to_counts[root.id] = defaultdict(int)
            actual_link_types_to_counts[root.id] = defaultdict(int)

            for link in root.get_links():
                expected_link_types_to_counts[root.id][link.rel] += 1

            for link in root.clone().get_links():
                actual_link_types_to_counts[root.id][link.rel] += 1

            for item in items:
                expected_link_types_to_counts[item.id] = defaultdict(int)
                actual_link_types_to_counts[item.id] = defaultdict(int)
                for link in item.get_links():
                    expected_link_types_to_counts[item.id][link.rel] += 1
                for link in item.get_links():
                    actual_link_types_to_counts[item.id][link.rel] += 1

        assert set(expected_link_types_to_counts.keys()) == set(
            actual_link_types_to_counts.keys()
        )

        for obj_id in actual_link_types_to_counts:
            expected_counts = expected_link_types_to_counts[obj_id]
            actual_counts = actual_link_types_to_counts[obj_id]
            assert set(expected_counts.keys()) == set(actual_counts.keys())
            for rel in expected_counts:
                assert actual_counts[rel] == expected_counts[rel], (
                    "Clone of {} has {} {} links, original has {}".format(
                        obj_id, actual_counts[rel], rel, expected_counts[rel]
                    )
                )

    def test_save_uses_previous_catalog_type(self) -> None:
        catalog = TestCases.case_1()
        assert catalog.catalog_type == CatalogType.SELF_CONTAINED
        with tempfile.TemporaryDirectory() as tmp_dir:
            catalog.normalize_hrefs(tmp_dir)
            href = catalog.self_href
            catalog.save()

            cat2 = pystac.Catalog.from_file(href)
            assert cat2.catalog_type == CatalogType.SELF_CONTAINED

    def test_save_to_provided_href(self) -> None:
        with tempfile.TemporaryDirectory() as tmp_dir:
            catalog = TestCases.case_1()
            href = "https://stac.test"
            folder = os.path.join(tmp_dir, "cat")
            catalog.normalize_hrefs(href)
            catalog.save(catalog_type=CatalogType.ABSOLUTE_PUBLISHED, dest_href=folder)

            catalog_path = os.path.join(folder, "catalog.json")
            assert os.path.exists(catalog_path)
            result_cat = Catalog.from_file(catalog_path)
            for link in result_cat.get_child_links():
                assert cast(str, link.target).startswith(href)

    def test_save_relative_published_no_self_links(self) -> None:
        with tempfile.TemporaryDirectory() as tmp_dir:
            catalog = TestCases.case_1()
            href = "https://stac.test"
            folder = os.path.join(tmp_dir, "cat")
            catalog.normalize_hrefs(href)
            catalog.save(catalog_type=CatalogType.RELATIVE_PUBLISHED, dest_href=folder)

            catalog_path = os.path.join(folder, "catalog.json")
            assert os.path.exists(catalog_path)
            result_cat = Catalog.from_file(catalog_path)

            # Check that Items do not have a self link
            # Since Item.from_dict automatically adds a self link, we need to look at
            # the JSON files themselves.
            stac_io = pystac.StacIO.default()

            for current_cat, _, __ in result_cat.walk():
                for item_link in current_cat.get_item_links():
                    item_dict = stac_io.read_json(item_link)
                    self_link = next(
                        (
                            link
                            for link in item_dict.get("links", [])
                            if link["rel"] == "self"
                        ),
                        None,
                    )
                    assert self_link is None

    def test_save_with_different_stac_io(self) -> None:
        catalog = Catalog.from_file(
            TestCases.get_path("data-files/catalogs/test-case-1/catalog.json")
        )
        stac_io = MockStacIO()
        with tempfile.TemporaryDirectory() as tmp_dir:
            catalog.normalize_hrefs(tmp_dir)
            catalog.save(
                catalog_type=CatalogType.ABSOLUTE_PUBLISHED,
                dest_href=tmp_dir,
                stac_io=stac_io,
            )

        hrefs = []
        for root, _, items in catalog.walk():
            hrefs.append(root.get_self_href())
            for item in items:
                hrefs.append(item.get_self_href())

        assert len(hrefs) == stac_io.mock.write_text.call_count
        for call_args_list in stac_io.mock.write_text.call_args_list:
            assert call_args_list[0][0] in hrefs

    def test_subcatalogs_saved_to_correct_path(self) -> None:
        with tempfile.TemporaryDirectory() as tmp_dir:
            catalog = TestCases.case_1()
            href = "https://stac.test"

            catalog.normalize_hrefs(href)
            catalog.save(catalog_type=CatalogType.ABSOLUTE_PUBLISHED, dest_href=tmp_dir)

            # Check the root catalog path
            expected_root_catalog_path = os.path.join(tmp_dir, "catalog.json")
            assert os.path.exists(expected_root_catalog_path), (
                f"{expected_root_catalog_path} does not exist."
            )
            assert os.path.isfile(expected_root_catalog_path), (
                f"{expected_root_catalog_path} is not a file."
            )

            # Check each child catalog
            for child_catalog in catalog.get_children():
                relative_path = make_relative_href(
                    child_catalog.self_href, catalog.self_href, start_is_dir=False
                )
                expected_child_path = make_absolute_href(
                    relative_path,
                    expected_root_catalog_path,
                    start_is_dir=False,
                )
                assert os.path.exists(expected_child_path), (
                    f"{expected_child_path} does not exist."
                )
                assert os.path.isfile(expected_child_path), (
                    f"{expected_child_path} is not a file."
                )

            # Check each item
            for item in catalog.get_items(recursive=True):
                relative_path = make_relative_href(
                    item.self_href, catalog.self_href, start_is_dir=False
                )
                expected_item_path = make_absolute_href(
                    relative_path,
                    expected_root_catalog_path,
                    start_is_dir=False,
                )
                assert os.path.exists(expected_item_path), (
                    f"{expected_item_path} does not exist."
                )
                assert os.path.isfile(expected_item_path), (
                    f"{expected_item_path} is not a file."
                )

    def test_clone_uses_previous_catalog_type(self) -> None:
        catalog = TestCases.case_1()
        assert catalog.catalog_type == CatalogType.SELF_CONTAINED
        clone = catalog.clone()
        assert clone.catalog_type == CatalogType.SELF_CONTAINED

    def test_normalize_hrefs_sets_all_hrefs(self) -> None:
        catalog = TestCases.case_1()
        catalog.normalize_hrefs("http://example.com")
        for root, _, items in catalog.walk():
            self_href = root.get_self_href()
            assert self_href is not None
            assert self_href.startswith("http://example.com")
            for link in root.links:
                if link.is_resolved():
                    target_href = cast(pystac.STACObject, link.target).self_href
                else:
                    target_href = link.absolute_href
                assert "http://example.com" in target_href, (
                    '[{}] {} does not contain "{}"'.format(
                        link.rel, target_href, "http://example.com"
                    )
                )
            for item in items:
                assert "http://example.com" in item.self_href

    def test_normalize_hrefs_makes_absolute_href(self) -> None:
        catalog = TestCases.case_1()
        catalog.normalize_hrefs("./relativepath")
        abspath = make_posix_style(os.path.abspath("./relativepath"))
        self_href = catalog.get_self_href()
        assert self_href is not None
        assert self_href.startswith(abspath)

    def test_normalize_hrefs_skip_unresolved(self) -> None:
        catalog = TestCases.case_1()
        catalog.normalize_hrefs("http://example.com", skip_unresolved=True)
        assert catalog.self_href.startswith("http://example.com")
        for link in catalog.links:
            if link.rel == "child" or link.rel == "item":
                assert not link.is_resolved()
        item = Item(
            "an-id",
            geometry=None,
            bbox=None,
            datetime=datetime.now(),
            properties={},
        )
        catalog.add_item(item, title="This is the test item")
        catalog.normalize_hrefs("http://example.com", skip_unresolved=True)
        for link in catalog.links:
            if link.title == "This is the test item":
                assert link.is_resolved()
                assert isinstance(link.target, pystac.STACObject)
                assert link.target.self_href.startswith("http://example.com")
            elif link.rel == "child" or link.rel == "item":
                assert not link.is_resolved()

    def test_save_unresolved(self) -> None:
        catalog = Catalog("an-id", "a description")
        item = Item(
            "an-id",
            geometry=None,
            bbox=None,
            datetime=datetime.now(),
            properties={},
        )
        catalog.add_item(item)
        catalog.add_link(pystac.Link("child", "/not/a/real/path/catalog.json"))
        with tempfile.TemporaryDirectory() as temporary_directory:
            catalog.set_self_href(os.path.join(temporary_directory, "catalog.json"))
            item.set_self_href(os.path.join(temporary_directory, "item.json"))
            catalog.save()
            assert len(os.listdir(temporary_directory)) == 2

        with tempfile.TemporaryDirectory() as temporary_directory:
            catalog.normalize_and_save(temporary_directory, skip_unresolved=True)
            assert len(os.listdir(temporary_directory)) == 2

        with tempfile.TemporaryDirectory() as temporary_directory:
            with pytest.raises(STACError, match="does not resolve to a STAC object"):
                catalog.normalize_and_save(temporary_directory, skip_unresolved=False)

    def test_generate_subcatalogs_works_with_custom_properties(self) -> None:
        catalog = TestCases.case_8()
        defaults = {"pl:item_type": "PlanetScope"}
        catalog.generate_subcatalogs(
            "${year}/${month}/${pl:item_type}", defaults=defaults
        )

        month_cat = catalog.get_child("8", recursive=True)
        assert month_cat is not None
        type_cats = {cat.id for cat in month_cat.get_children()}

        assert type_cats == {"PSScene4Band", "SkySatScene", "PlanetScope"}

    def test_generate_subcatalogs_does_not_change_item_count(self) -> None:
        catalog = TestCases.case_7()

        item_counts = {
            cat.id: len(list(cat.get_items(recursive=True)))
            for cat in catalog.get_children()
        }

        catalog.generate_subcatalogs("${year}/${day}")

        with tempfile.TemporaryDirectory() as tmp_dir:
            catalog.normalize_hrefs(tmp_dir)
            catalog.save(pystac.CatalogType.SELF_CONTAINED)

            cat2 = pystac.Catalog.from_file(os.path.join(tmp_dir, "catalog.json"))
            for child in cat2.get_children():
                actual = len(list(child.get_items(recursive=True)))
                expected = item_counts[child.id]
                assert actual == expected, f" for child '{child.id}'"

    def test_generate_subcatalogs_merge_template_elements(self) -> None:
        catalog = Catalog(id="test", description="Test")
        item_properties = [
            dict(property1=p1, property2=p2) for p1 in ("A", "B") for p2 in (1, 2)
        ]
        for ni, properties in enumerate(item_properties):
            catalog.add_item(
                Item(
                    id=f"item{ni}",
                    geometry=ARBITRARY_GEOM,
                    bbox=ARBITRARY_BBOX,
                    datetime=datetime.now(timezone.utc),
                    properties=properties,
                )
            )
        result = catalog.generate_subcatalogs("${property1}_${property2}")

        actual_subcats = {cat.id for cat in result}
        expected_subcats = {
            "{}_{}".format(d["property1"], d["property2"]) for d in item_properties
        }
        assert len(result) == len(expected_subcats)
        assert actual_subcats == expected_subcats

    def test_generate_subcatalogs_can_be_applied_multiple_times(self) -> None:
        catalog = TestCases.case_8()

        _ = catalog.generate_subcatalogs("${year}/${month}")
        catalog.normalize_hrefs("/tmp")
        expected_hrefs = {
            item.id: item.get_self_href() for item in catalog.get_items(recursive=True)
        }

        result = catalog.generate_subcatalogs("${year}/${month}")
        assert len(result) == 0
        catalog.normalize_hrefs("/tmp")
        for item in catalog.get_items(recursive=True):
            assert item.get_self_href() == expected_hrefs[item.id], (
                f" for item '{item.id}'"
            )

    def test_generate_subcatalogs_works_after_adding_more_items(self) -> None:
        catalog = Catalog(id="test", description="Test")
        properties = dict(property1="A", property2=1)
        catalog.add_item(
            Item(
                id="item1",
                geometry=ARBITRARY_GEOM,
                bbox=ARBITRARY_BBOX,
                datetime=datetime.now(timezone.utc),
                properties=properties,
            )
        )
        catalog.generate_subcatalogs("${property1}/${property2}")
        catalog.add_item(
            Item(
                id="item2",
                geometry=ARBITRARY_GEOM,
                bbox=ARBITRARY_BBOX,
                datetime=datetime.now(timezone.utc),
                properties=properties,
            )
        )
        catalog.generate_subcatalogs("${property1}/${property2}")

        catalog.normalize_hrefs("/tmp")
        item1 = next(catalog.get_items("item1", recursive=True))
        item1_parent = item1.get_parent()
        assert item1_parent is not None
        item2 = next(catalog.get_items("item2", recursive=True))
        item2_parent = item2.get_parent()
        assert item2_parent is not None
        assert item1_parent.get_self_href() == item2_parent.get_self_href()

    def test_generate_subcatalogs_works_for_branched_subcatalogs(self) -> None:
        catalog = Catalog(id="test", description="Test")
        item_properties = [
            dict(property1="A", property2=1, property3="i"),  # add 3 subcats
            dict(property1="A", property2=1, property3="j"),  # add 1 more
            dict(property1="A", property2=2, property3="i"),  # add 2 more
            dict(property1="B", property2=1, property3="i"),  # add 3 more
        ]
        for ni, properties in enumerate(item_properties):
            catalog.add_item(
                Item(
                    id=f"item{ni}",
                    geometry=ARBITRARY_GEOM,
                    bbox=ARBITRARY_BBOX,
                    datetime=datetime.now(timezone.utc),
                    properties=properties,
                )
            )
        result = catalog.generate_subcatalogs("${property1}/${property2}/${property3}")
        assert len(result) == 9

        actual_subcats = {cat.id for cat in result}
        expected_subcats = {"A", "B", "1", "2", "i", "j"}
        assert actual_subcats == expected_subcats

    def test_generate_subcatalogs_works_for_subcatalogs_with_same_ids(self) -> None:
        catalog = Catalog(id="test", description="Test")
        item_properties = [
            dict(property1=1, property2=1),  # add 2 subcats
            dict(property1=1, property2=2),  # add 1 more
            dict(property1=2, property2=1),  # add 2 more
            dict(property1=2, property2=2),  # add 1 more
        ]
        for ni, properties in enumerate(item_properties):
            catalog.add_item(
                Item(
                    id=f"item{ni}",
                    geometry=ARBITRARY_GEOM,
                    bbox=ARBITRARY_BBOX,
                    datetime=datetime.now(timezone.utc),
                    properties=properties,
                )
            )

        result = catalog.generate_subcatalogs("${property1}/${property2}")
        assert len(result) == 6

        catalog.normalize_hrefs("/")

        for item in catalog.get_items(recursive=True):
            item_parent = item.get_parent()
            assert item_parent is not None
            parent_href = item_parent.self_href
            path_to_parent, _ = os.path.split(parent_href)
            subcats = list(
                Path(path_to_parent).parts[1:]
            )  # Skip drive letter if present (Windows)

            assert len(subcats) == 2, f" for item '{item.id}'"

    def test_map_items(self) -> None:
        def item_mapper(item: pystac.Item) -> pystac.Item:
            item.properties["ITEM_MAPPER"] = "YEP"
            return item

        with tempfile.TemporaryDirectory() as tmp_dir:
            catalog = TestCases.case_1()

            new_cat = catalog.map_items(item_mapper)

            new_cat.normalize_hrefs(os.path.join(tmp_dir, "cat"))
            new_cat.save(catalog_type=CatalogType.ABSOLUTE_PUBLISHED)

            result_cat = Catalog.from_file(os.path.join(tmp_dir, "cat", "catalog.json"))

            for item in result_cat.get_items(recursive=True):
                assert "ITEM_MAPPER" in item.properties

            for item in catalog.get_items(recursive=True):
                assert "ITEM_MAPPER" not in item.properties

    def test_map_items_multiple(self) -> None:
        def item_mapper(item: pystac.Item) -> list[pystac.Item]:
            item2 = item.clone()
            item2.id = item2.id + "_2"
            item.properties["ITEM_MAPPER_1"] = "YEP"
            item2.properties["ITEM_MAPPER_2"] = "YEP"
            return [item, item2]

        with tempfile.TemporaryDirectory() as tmp_dir:
            catalog = TestCases.case_1()
            catalog_items = list(catalog.get_items(recursive=True))

            new_cat = catalog.map_items(item_mapper)

            new_cat.normalize_hrefs(os.path.join(tmp_dir, "cat"))
            new_cat.save(catalog_type=CatalogType.ABSOLUTE_PUBLISHED)

            result_cat = Catalog.from_file(os.path.join(tmp_dir, "cat", "catalog.json"))
            result_items = list(result_cat.get_items(recursive=True))

            assert len(catalog_items) * 2 == len(result_items)

            ones, twos = 0, 0
            for item in result_items:
                assert ("ITEM_MAPPER_1" in item.properties) or (
                    "ITEM_MAPPER_2" in item.properties
                )
                if "ITEM_MAPPER_1" in item.properties:
                    ones += 1

                if "ITEM_MAPPER_2" in item.properties:
                    twos += 1

            assert ones == twos

            for item in catalog.get_items(recursive=True):
                assert ("ITEM_MAPPER_1" not in item.properties) and (
                    "ITEM_MAPPER_2" not in item.properties
                )

    def test_map_items_multiple_2(self) -> None:
        catalog = Catalog(id="test-1", description="Test1")
        item1 = Item(
            id="item1",
            geometry=ARBITRARY_GEOM,
            bbox=ARBITRARY_BBOX,
            datetime=datetime.now(timezone.utc),
            properties={},
        )
        item1.add_asset("ortho", Asset(href="/some/ortho.tif"))
        catalog.add_item(item1)
        kitten = Catalog(id="test-kitten", description="A cuter version of catalog")
        catalog.add_child(kitten)
        item2 = Item(
            id="item2",
            geometry=ARBITRARY_GEOM,
            bbox=ARBITRARY_BBOX,
            datetime=datetime.now(timezone.utc),
            properties={},
        )
        item2.add_asset("ortho", Asset(href="/some/other/ortho.tif"))
        kitten.add_item(item2)

        def modify_item_title(item: pystac.Item) -> pystac.Item:
            item.properties["title"] = "Some title"
            return item

        def duplicate_item(item: pystac.Item) -> list[pystac.Item]:
            duplicated_item = item.clone()
            duplicated_item.id += "-duplicated"
            return [item, duplicated_item]

        c = catalog.map_items(modify_item_title)
        c = c.map_items(duplicate_item)
        new_catalog = c

        items = new_catalog.get_items(recursive=True)
        assert len(list(items)) == 4

    def test_map_assets_single(self) -> None:
        changed_asset = "d43bead8-e3f8-4c51-95d6-e24e750a402b"

        def asset_mapper(key: str, asset: pystac.Asset) -> pystac.Asset:
            if key == changed_asset:
                asset.title = "NEW TITLE"

            return asset

        with tempfile.TemporaryDirectory() as tmp_dir:
            catalog = TestCases.case_2()

            new_cat = catalog.map_assets(asset_mapper)

            new_cat.normalize_hrefs(os.path.join(tmp_dir, "cat"))
            new_cat.save(catalog_type=CatalogType.ABSOLUTE_PUBLISHED)

            result_cat = Catalog.from_file(os.path.join(tmp_dir, "cat", "catalog.json"))

            found = False
            for item in result_cat.get_items(recursive=True):
                for key, asset in item.assets.items():
                    if key == changed_asset:
                        found = True
                        assert asset.title == "NEW TITLE"
                    else:
                        assert asset.title != "NEW TITLE"
            assert found

    def test_map_assets_tup(self) -> None:
        changed_assets: list[str] = []

        def asset_mapper(
            key: str, asset: pystac.Asset
        ) -> pystac.Asset | tuple[str, pystac.Asset]:
            if asset.media_type and "geotiff" in asset.media_type:
                asset.title = "NEW TITLE"
                changed_assets.append(key)
                return (f"{key}-modified", asset)
            else:
                return asset

        with tempfile.TemporaryDirectory() as tmp_dir:
            catalog = TestCases.case_2()

            new_cat = catalog.map_assets(asset_mapper)

            new_cat.normalize_hrefs(os.path.join(tmp_dir, "cat"))
            new_cat.save(catalog_type=CatalogType.ABSOLUTE_PUBLISHED)

            result_cat = Catalog.from_file(os.path.join(tmp_dir, "cat", "catalog.json"))

            found = False
            not_found = False
            for item in result_cat.get_items(recursive=True):
                for key, asset in item.assets.items():
                    if key.replace("-modified", "") in changed_assets:
                        found = True
                        assert asset.title == "NEW TITLE"
                    else:
                        not_found = True
                        assert asset.title != "NEW TITLE"

            assert found
            assert not_found

    def test_map_assets_multi(self) -> None:
        changed_assets = []

        def asset_mapper(
            key: str, asset: pystac.Asset
        ) -> pystac.Asset | dict[str, pystac.Asset]:
            if asset.media_type and "geotiff" in asset.media_type:
                changed_assets.append(key)
                mod1 = asset.clone()
                mod1.title = "NEW TITLE 1"
                mod2 = asset.clone()
                mod2.title = "NEW TITLE 2"
                return {f"{key}-mod-1": mod1, f"{key}-mod-2": mod2}
            else:
                return asset

        with tempfile.TemporaryDirectory() as tmp_dir:
            catalog = TestCases.case_2()

            new_cat = catalog.map_assets(asset_mapper)

            new_cat.normalize_hrefs(os.path.join(tmp_dir, "cat"))
            new_cat.save(catalog_type=CatalogType.ABSOLUTE_PUBLISHED)

            result_cat = Catalog.from_file(os.path.join(tmp_dir, "cat", "catalog.json"))

            found1 = False
            found2 = False
            not_found = False
            for item in result_cat.get_items(recursive=True):
                for key, asset in item.assets.items():
                    if key.replace("-mod-1", "") in changed_assets:
                        found1 = True
                        assert asset.title == "NEW TITLE 1"
                    elif key.replace("-mod-2", "") in changed_assets:
                        found2 = True
                        assert asset.title == "NEW TITLE 2"
                    else:
                        not_found = True
                        assert asset.title != "NEW TITLE"

            assert found1
            assert found2
            assert not_found

    def test_make_all_asset_hrefs_absolute(self) -> None:
        cat = TestCases.case_2()
        cat.make_all_asset_hrefs_absolute()
        ID = "cf73ec1a-d790-4b59-b077-e101738571ed"
        item = next(cat.get_items(ID, recursive=True))
        href = item.assets[ID].href
        assert is_absolute_href(href)

    def test_make_all_asset_hrefs_relative(self) -> None:
        cat = TestCases.case_2()
        ID = "cf73ec1a-d790-4b59-b077-e101738571ed"
        item = next(cat.get_items(ID, recursive=True))
        asset = item.assets[ID]
        original_href = asset.href
        cat.make_all_asset_hrefs_absolute()

        assert is_absolute_href(asset.href)

        cat.make_all_asset_hrefs_relative()

        assert not is_absolute_href(asset.href)
        assert asset.href == original_href

    @pytest.mark.parametrize("catalog", TestCases.all_test_catalogs())
    def test_make_all_links_relative_or_absolute(self, catalog: Catalog) -> None:
        def check_all_relative(cat: Catalog) -> None:
            for root, catalogs, items in cat.walk():
                for link in root.links:
                    if link.rel in HIERARCHICAL_LINKS:
                        assert not is_absolute_href(link.href)
                for item in items:
                    for link in item.links:
                        if link.rel in HIERARCHICAL_LINKS:
                            assert not is_absolute_href(link.href)

        def check_all_absolute(cat: Catalog) -> None:
            for root, catalogs, items in cat.walk():
                for link in root.links:
                    assert is_absolute_href(link.href)
                for item in items:
                    for link in item.links:
                        assert is_absolute_href(link.href)

        with tempfile.TemporaryDirectory() as tmp_dir:
            c2 = catalog.full_copy()
            c2.normalize_hrefs(tmp_dir)
            c2.catalog_type = CatalogType.RELATIVE_PUBLISHED
            check_all_relative(c2)
            c2.catalog_type = CatalogType.ABSOLUTE_PUBLISHED
            check_all_absolute(c2)

    @pytest.mark.block_network()
    def test_self_contained_catalog_collection_item_links(self) -> None:
        """See issue https://github.com/stac-utils/pystac/issues/657"""
        with tempfile.TemporaryDirectory() as tmp_dir:
            catalog = pystac.Catalog(
                id="catalog-issue-657", description="catalog-issue-657"
            )
            collection = pystac.Collection(
                "collection-issue-657",
                "collection-issue-657",
                pystac.Extent(
                    spatial=pystac.SpatialExtent([[-180.0, -90.0, 180.0, 90.0]]),
                    temporal=pystac.TemporalExtent([[datetime(2021, 11, 1), None]]),
                ),
                license="other",
            )

            item = pystac.Item(
                id="item-issue-657",
                stac_extensions=[],
                geometry=ARBITRARY_GEOM,
                bbox=ARBITRARY_BBOX,
                datetime=datetime(2021, 11, 1),
                properties={},
            )

            collection.add_item(item)
            catalog.add_child(collection)

            catalog.normalize_hrefs(tmp_dir)
            catalog.validate_all()

            catalog.save(catalog_type=CatalogType.SELF_CONTAINED)
            with open(
                f"{tmp_dir}/collection-issue-657/item-issue-657/item-issue-657.json"
            ) as f:
                item_json = json.load(f)

            for link in item_json["links"]:
                # self links are always absolute
                if link["rel"] == "self":
                    continue

                href = link["href"]
                assert not is_absolute_href(href), (
                    f"Link with rel={link['rel']} is absolute!"
                )

    def test_full_copy_and_normalize_works_with_created_stac(self) -> None:
        cat = TestCases.case_3()
        cat_copy = cat.full_copy()
        cat_copy.normalize_hrefs("http://example.com")
        for root, catalogs, items in cat_copy.walk():
            for link in root.links:
                if link.rel != "self":
                    assert link.target is not None
            for item in items:
                for link in item.links:
                    if link.rel != "self":
                        assert link.get_href() is not None

    def test_extra_fields(self) -> None:
        catalog = TestCases.case_1()

        catalog.extra_fields["custom_field"] = "Special content"

        with tempfile.TemporaryDirectory() as tmp_dir:
            p = os.path.join(tmp_dir, "catalog.json")
            catalog.save_object(include_self_link=False, dest_href=p)
            with open(p) as f:
                cat_json = json.load(f)
            assert "custom_field" in cat_json
            assert cat_json["custom_field"] == "Special content"

            read_cat = pystac.Catalog.from_file(p)
            assert "custom_field" in read_cat.extra_fields
            assert read_cat.extra_fields["custom_field"] == "Special content"

    @pytest.mark.parametrize("cat", TestCases.all_test_catalogs())
    @pytest.mark.vcr()
    def test_validate_all(self, cat: Catalog) -> None:
        # If hrefs are not set, it will fail validation.
        if cat.get_self_href() is None:
            cat.normalize_hrefs("/tmp")
        cat.validate_all()

    @pytest.mark.vcr()
    def test_validate_all_invalid(self) -> None:
        # Make one invalid, write it off, read it in, ensure it throws
        cat = TestCases.case_1()
        item = next(cat.get_items("area-1-1-labels", recursive=True))
        item.geometry = {"type": "INVALID", "coordinates": "NONE"}
        with tempfile.TemporaryDirectory() as tmp_dir:
            cat.normalize_hrefs(tmp_dir)
            cat.save(catalog_type=pystac.CatalogType.SELF_CONTAINED)

            cat2 = pystac.Catalog.from_file(os.path.join(tmp_dir, "catalog.json"))

            with pytest.raises(pystac.STACValidationError):
                cat2.validate_all()

    def test_set_hrefs_manually(self) -> None:
        catalog = TestCases.case_1()

        # Modify the datetimes
        year = 2004
        month = 2
        for item in catalog.get_items(recursive=True):
            assert item.datetime is not None
            item.datetime = item.datetime.replace(year=year, month=month)
            year += 1
            month += 1

        with tempfile.TemporaryDirectory() as tmp_dir:
            for root, _, items in catalog.walk():
                # Set root's HREF based off the parent
                parent = root.get_parent()
                if parent is None:
                    root_dir = tmp_dir
                else:
                    d = os.path.dirname(parent.self_href)
                    root_dir = os.path.join(d, root.id)
                root_href = os.path.join(root_dir, root.DEFAULT_FILE_NAME)
                root.set_self_href(root_href)

                # Set each item's HREF based on it's datetime
                for item in items:
                    assert item.datetime is not None
                    item_href = "{}/{}-{}/{}.json".format(
                        root_dir, item.datetime.year, item.datetime.month, item.id
                    )
                    item.set_self_href(item_href)

            catalog.save(catalog_type=CatalogType.SELF_CONTAINED)

            read_catalog = Catalog.from_file(os.path.join(tmp_dir, "catalog.json"))

            for root, _, items in read_catalog.walk():
                parent = root.get_parent()
                if parent is None:
                    assert root.get_self_href() == make_posix_style(
                        os.path.join(tmp_dir, "catalog.json")
                    )
                else:
                    d = os.path.dirname(parent.self_href)
                    assert root.get_self_href() == make_posix_style(
                        os.path.join(d, root.id, root.DEFAULT_FILE_NAME)
                    )
                for item in items:
                    assert item.datetime is not None
                    end = posixpath.join(
                        f"{item.datetime.year}-{item.datetime.month}",
                        f"{item.id}.json",
                    )
                    self_href = item.get_self_href()
                    assert self_href is not None
                    assert self_href.endswith(end), "{} does not end with {}".format(
                        self_href, end
                    )

    @pytest.mark.parametrize("cat", TestCases.all_test_catalogs())
    def test_collections_cache_correctly(self, cat: Catalog) -> None:
        mock_io = MockStacIO()
        cat._stac_io = mock_io
        expected_collection_reads = set()
        for root, _, items in cat.walk():
            if isinstance(root, Collection) and root != cat:
                expected_collection_reads.add(root.get_self_href())

            # Iterate over items to make sure they are read
            assert list(items) is not None

        call_uris: list[Any] = [
            call[0][0]
            for call in mock_io.mock.read_text.call_args_list
            if call[0][0] in expected_collection_reads
        ]

        for collection_uri in expected_collection_reads:
            calls = len([x for x in call_uris if x == collection_uri])
            assert calls == 1, "{} was read {} times instead of once!".format(
                collection_uri, calls
            )

    def test_reading_iterating_and_writing_works_as_expected(self) -> None:
        """Test case to cover issue #88"""
        stac_uri = TestCases.get_path("data-files/catalogs/test-case-6/catalog.json")
        cat = Catalog.from_file(stac_uri)

        # Iterate over the items. This was causing failure in
        # the later iterations as per issue #88
        for item in cat.get_items(recursive=True):
            pass

        with tempfile.TemporaryDirectory() as tmp_dir:
            new_stac_uri = os.path.join(tmp_dir, "test-case-6")
            cat.normalize_hrefs(new_stac_uri)
            cat.save(catalog_type=CatalogType.SELF_CONTAINED)

            # Open the local copy and iterate over it.
            cat2 = Catalog.from_file(os.path.join(new_stac_uri, "catalog.json"))

            for item in cat2.get_items(recursive=True):
                # Iterate again over the items. This would fail in #88
                pass

    def test_get_children_cbers(self) -> None:
        cat = TestCases.case_6()
        assert len(list(cat.get_children())) == 4

    def test_resolve_planet(self) -> None:
        """Test against a bug that caused infinite recursion during link resolution"""
        cat = TestCases.case_8()
        for root, _, items in cat.walk():
            for item in items:
                item.resolve_links()
            root.resolve_links()

    def test_handles_children_with_same_id(self) -> None:
        # This catalog has the root and child collection share an ID.
        cat = pystac.Catalog.from_file(
            TestCases.get_path("data-files/invalid/shared-id/catalog.json")
        )
        items = list(cat.get_items(recursive=True))

        assert len(items) == 1

    def test_catalog_with_href_caches_by_href(self) -> None:
        cat = TestCases.case_1()
        cache = cat._resolved_objects

        # Since all of our STAC objects have HREFs, everything should be
        # cached only by HREF
        assert len(cache.id_keys_to_objects) == 0

    def test_from_invalid_dict_raises_exception(self) -> None:
        stac_io = pystac.StacIO.default()
        collection_dict = stac_io.read_json(
            TestCases.get_path("data-files/collections/multi-extent.json")
        )
        with pytest.raises(pystac.STACTypeError):
            _ = pystac.Catalog.from_dict(collection_dict)

    def test_get_collections(self) -> None:
        catalog = TestCases.case_2()
        collections = list(catalog.get_collections())

        assert len(collections) > 0
        assert all(isinstance(c, pystac.Collection) for c in collections)

    def test_get_all_collections(self) -> None:
        catalog = TestCases.case_1()
        all_collections = list(catalog.get_all_collections())

        assert len(all_collections) == 4
        assert all(isinstance(c, pystac.Collection) for c in all_collections)

    def test_get_single_links_media_type(self) -> None:
        catalog = TestCases.case_1()

        catalog.links.append(
            pystac.Link(rel="search", target="./search.html", media_type="text/html")
        )
        catalog.links.append(
            pystac.Link(
                rel="search", target="./search.json", media_type="application/geo+json"
            )
        )

        html_link = catalog.get_single_link(rel="search")
        assert html_link is not None
        assert html_link.href == "./search.html"
        html_link = catalog.get_single_link(media_type="text/html")
        assert html_link is not None
        assert html_link.href == "./search.html"
        json_link = catalog.get_single_link(
            rel="search", media_type="application/geo+json"
        )
        assert json_link is not None
        assert json_link.href == "./search.json"
        no_link = catalog.get_single_link(rel="via")
        assert no_link is None
        first_link = catalog.get_single_link()
        assert first_link is not None
        assert first_link.rel == "self"

    def test_get_links(self) -> None:
        catalog = TestCases.case_1()

        catalog.links.append(
            pystac.Link(rel="search", target="./search.html", media_type="text/html")
        )
        catalog.links.append(
            pystac.Link(
                rel="search", target="./search.json", media_type="application/geo+json"
            )
        )
        assert (
            len(catalog.get_links(rel="search", media_type="application/geo+json")) == 1
        )
        assert len(catalog.get_links(media_type="text/html")) == 1
        assert (
            len(catalog.get_links(media_type=["text/html", "application/geo+json"]))
            == 2
        )
        assert len(catalog.get_links(rel="search")) == 2
        assert len(catalog.get_links(rel="via")) == 0
        assert len(catalog.get_links()) == 6

    def test_to_dict_no_self_href(self) -> None:
        catalog = Catalog(id="an-id", description="A test Catalog")
        d = catalog.to_dict(include_self_link=False)
        Catalog.from_dict(d)


class TestFullCopy:
    def check_link(self, link: pystac.Link, tag: str) -> None:
        if link.is_resolved():
            target_href: str = cast(pystac.STACObject, link.target).self_href
        else:
            target_href = str(link.target)
        assert tag in target_href, '[{}] {} does not contain "{}"'.format(
            link.rel, target_href, tag
        )

    def check_item(self, item: Item, tag: str) -> None:
        for link in item.links:
            self.check_link(link, tag)

    def check_catalog(self, c: Catalog, tag: str) -> None:
        assert len(c.get_links("root")) == 1, f"Failure for catalog: {c}"

        for link in c.links:
            self.check_link(link, tag)

        for child in c.get_children():
            self.check_catalog(child, tag)

        for item in c.get_items():
            self.check_item(item, tag)

    def test_full_copy_1(self) -> None:
        with tempfile.TemporaryDirectory() as tmp_dir:
            cat = Catalog(id="test", description="test catalog")

            item = Item(
                id="test_item",
                geometry=ARBITRARY_GEOM,
                bbox=ARBITRARY_BBOX,
                datetime=datetime.now(timezone.utc),
                properties={},
            )

            cat.add_item(item)

            cat.normalize_hrefs(os.path.join(tmp_dir, "catalog-full-copy-1-source"))
            cat2 = cat.full_copy()
            cat2.normalize_hrefs(os.path.join(tmp_dir, "catalog-full-copy-1-dest"))

            self.check_catalog(cat, "source")
            self.check_catalog(cat2, "dest")

    def test_full_copy_2(self) -> None:
        with tempfile.TemporaryDirectory() as tmp_dir:
            cat = Catalog(id="test", description="test catalog")
            image_item = Item(
                id="Imagery",
                geometry=ARBITRARY_GEOM,
                bbox=ARBITRARY_BBOX,
                datetime=datetime.now(timezone.utc),
                properties={},
            )
            for key in ["ortho", "dsm"]:
                image_item.add_asset(
                    key,
                    Asset(href=f"some/{key}.tif", media_type=MediaType.GEOTIFF),
                )

            label_item = Item(
                id="Labels",
                geometry=ARBITRARY_GEOM,
                bbox=ARBITRARY_BBOX,
                datetime=datetime.now(timezone.utc),
                properties={},
            )
            cat.add_items([image_item, label_item])

            cat.normalize_hrefs(os.path.join(tmp_dir, "catalog-full-copy-2-source"))
            cat.save(catalog_type=CatalogType.ABSOLUTE_PUBLISHED)
            cat2 = cat.full_copy()
            cat2.normalize_hrefs(os.path.join(tmp_dir, "catalog-full-copy-2-dest"))
            cat2.save(catalog_type=CatalogType.ABSOLUTE_PUBLISHED)

            self.check_catalog(cat, "source")
            self.check_catalog(cat2, "dest")

    def test_full_copy_3(self) -> None:
        with tempfile.TemporaryDirectory() as tmp_dir:
            root_cat = TestCases.case_1()
            root_cat.normalize_hrefs(
                os.path.join(tmp_dir, "catalog-full-copy-3-source")
            )
            root_cat.save(catalog_type=CatalogType.ABSOLUTE_PUBLISHED)
            cat2 = root_cat.full_copy()
            cat2.normalize_hrefs(os.path.join(tmp_dir, "catalog-full-copy-3-dest"))
            cat2.save(catalog_type=CatalogType.ABSOLUTE_PUBLISHED)

            self.check_catalog(root_cat, "source")
            self.check_catalog(cat2, "dest")

    def test_full_copy_4(self) -> None:
        with tempfile.TemporaryDirectory() as tmp_dir:
            root_cat = TestCases.case_2()
            root_cat.normalize_hrefs(
                os.path.join(tmp_dir, "catalog-full-copy-4-source")
            )
            root_cat.save(catalog_type=CatalogType.ABSOLUTE_PUBLISHED)
            cat2 = root_cat.full_copy()
            cat2.normalize_hrefs(os.path.join(tmp_dir, "catalog-full-copy-4-dest"))
            cat2.save(catalog_type=CatalogType.ABSOLUTE_PUBLISHED)

            self.check_catalog(root_cat, "source")
            self.check_catalog(cat2, "dest")

            # Check that the relative asset link was saved correctly in the copy.
            ID = "cf73ec1a-d790-4b59-b077-e101738571ed"
            item = next(cat2.get_items(ID, recursive=True))
            href = item.assets[ID].get_absolute_href()
            assert href is not None
            assert os.path.exists(href)


class TestCatalogSubClass:
    """This tests cases related to creating classes inheriting from pystac.Catalog to
    ensure that inheritance, class methods, etc. function as expected."""

    case_1 = TestCases.get_path("data-files/catalogs/test-case-1/catalog.json")

    class BasicCustomCatalog(pystac.Catalog):
        def get_items(self) -> Iterator[Item]:  # type: ignore
            # This get_items does not have the `recursive` kwarg. This mimics
            # the current state of pystac-client and is intended to test
            # backwards compatibility of inherited classes
            return super().get_items()

    def test_from_dict_returns_subclass(self) -> None:
        self.stac_io = pystac.StacIO.default()
        catalog_dict = self.stac_io.read_json(self.case_1)
        custom_catalog = self.BasicCustomCatalog.from_dict(catalog_dict)
        assert isinstance(custom_catalog, self.BasicCustomCatalog)

    def test_from_file_returns_subclass(self) -> None:
        custom_catalog = self.BasicCustomCatalog.from_file(self.case_1)
        assert isinstance(custom_catalog, self.BasicCustomCatalog)

    def test_clone(self) -> None:
        custom_catalog = self.BasicCustomCatalog.from_file(self.case_1)
        cloned_catalog = custom_catalog.clone()
        assert isinstance(cloned_catalog, self.BasicCustomCatalog)

    def test_get_all_items_works(self) -> None:
        custom_catalog = self.BasicCustomCatalog.from_file(self.case_1)
        cloned_catalog = custom_catalog.clone()
        with pytest.warns(DeprecationWarning):
            cloned_catalog.get_all_items()

    def test_get_item_works(self) -> None:
        custom_catalog = self.BasicCustomCatalog.from_file(self.case_1)
        cloned_catalog = custom_catalog.clone()
        with pytest.warns(DeprecationWarning):
            cloned_catalog.get_item("area-1-1-imagery")


def test_custom_catalog_from_dict(catalog: Catalog) -> None:
    # https://github.com/stac-utils/pystac/issues/862
    class CustomCatalog(Catalog):
        @classmethod
        def from_dict(
            cls,
            d: dict[str, Any],
            href: str | None = None,
            root: Catalog | None = None,
            migrate: bool = True,
            preserve_dict: bool = True,
        ) -> CustomCatalog:
            return super().from_dict(d)

    _ = CustomCatalog.from_dict(catalog.to_dict())


@pytest.mark.parametrize("add_canonical", (True, False))
def test_remove_hierarchical_links(
    test_case_1_catalog: Catalog, add_canonical: bool
) -> None:
    test_case_1_catalog.remove_hierarchical_links(add_canonical=add_canonical)
    for link in test_case_1_catalog.links:
        assert not link.is_hierarchical()
    assert bool(test_case_1_catalog.get_single_link("canonical")) == add_canonical


def test_fully_resolve(tmp_path: Path, test_case_1_catalog: Catalog) -> None:
    test_case_1_catalog.save(dest_href=str(tmp_path / "before"))
    assert len(list((tmp_path / "before").glob("**/*.json"))) == 1
    test_case_1_catalog.fully_resolve()
    test_case_1_catalog.save(dest_href=str(tmp_path / "after"))
    assert len(list((tmp_path / "after").glob("**/*.json"))) == 15


def test_get_items_with_multiple_ids(test_case_1_catalog: Catalog) -> None:
    cat = test_case_1_catalog
    items = cat.get_items("area-2-1-imagery", "area-1-1-labels", recursive=True)
    assert len(list(items)) == 2


@pytest.mark.vcr()
def test_validate_all_with_max_n(test_case_1_catalog: Catalog) -> None:
    cat = test_case_1_catalog
    assert cat.validate_all() == 8
    assert cat.validate_all(max_items=6) == 6
    assert cat.validate_all(max_items=1) == 1


@pytest.mark.vcr()
def test_validate_all_with_recusive_off(test_case_1_catalog: Catalog) -> None:
    cat = test_case_1_catalog
    assert cat.validate_all() == 8
    assert cat.validate_all(recursive=False) == 0


@pytest.fixture
def nested_catalog() -> pystac.Catalog:
    """
    Structure:

    ├── catalog.json
    ├── products
    │   ├── catalog.json
    │   └── product_a
    │       └── catalog.json
    └── variables
        ├── catalog.json
        └── variable_a
            └── catalog.json
                └── variable_a_1
                    └── collection.json
    """
    root = pystac.Catalog("root", "root")
    variables = pystac.Catalog("variables", "variables")
    products = pystac.Catalog("products", "products")

    root.add_child(variables)
    root.add_child(products)

    variable_a = pystac.Catalog("variable_a", "variable_a")
    product_a = pystac.Catalog("product_a", "product_a")

    variables.add_child(variable_a)
    products.add_child(product_a)

    variable_a_1 = pystac.Collection(
        "variable_a_1", "variable_a_1", extent=ARBITRARY_EXTENT
    )
    variable_a.add_child(variable_a_1)

    return root


def test_get_all_collections_deeply_nested(nested_catalog: pystac.Catalog) -> None:
    catalog = nested_catalog
    all_collections = list(catalog.get_all_collections())

    assert len(all_collections) == 1
    assert all(isinstance(c, pystac.Collection) for c in all_collections)


def test_set_parent_false_stores_in_proper_place_on_normalize_and_save(
    nested_catalog: pystac.Catalog, tmp_path: Path
) -> None:
    root = nested_catalog
    product_a = next(root.get_child("products").get_children())  # type: ignore
    variable_a = next(root.get_child("variables").get_children())  # type: ignore

    variable_a.add_child(product_a, set_parent=False)

    root.normalize_and_save(
        root_href=str(tmp_path), catalog_type=pystac.CatalogType.ABSOLUTE_PUBLISHED
    )
    assert (tmp_path / "products" / "product_a").exists()
    assert not (tmp_path / "variables" / "variable_a" / "product_a").exists()


def test_set_parent_false_stores_in_proper_place_on_save(
    nested_catalog: pystac.Catalog, tmp_path: Path
) -> None:
    nested_catalog.normalize_and_save(
        root_href=str(tmp_path), catalog_type=pystac.CatalogType.ABSOLUTE_PUBLISHED
    )
    root = pystac.Catalog.from_file(tmp_path / "catalog.json")
    product_a = next(root.get_child("products").get_children())  # type: ignore
    variable_a = next(root.get_child("variables").get_children())  # type: ignore

    variable_a.add_child(product_a, set_parent=False)

    root.save(pystac.CatalogType.SELF_CONTAINED, dest_href=str(tmp_path))

    assert (tmp_path / "products" / "product_a").exists()
    assert not (tmp_path / "variables" / "variable_a" / "product_a").exists()


BEST_PRACTICE_CATALOG_TEMPLATE = "{id}"
BEST_PRACTICE_ITEM_TEMPLATE = "{id}"
TEST_CATALOG_TEMPLATE = "cat/${id}/${description}"
TEST_ITEM_TEMPLATE = "cat/items/${id}"
STRATEGY = TemplateLayoutStrategy(
    catalog_template=TEST_CATALOG_TEMPLATE, item_template=TEST_ITEM_TEMPLATE
)


@pytest.mark.parametrize(
    "root_strategy,sub_strategy,provided_root_strategy,provided_sub_strategy,root_template,sub_template",
    [
        (
            None,
            None,
            None,
            None,
            BEST_PRACTICE_CATALOG_TEMPLATE,
            BEST_PRACTICE_CATALOG_TEMPLATE,
        ),
        (STRATEGY, None, None, None, TEST_CATALOG_TEMPLATE, TEST_CATALOG_TEMPLATE),
        (
            STRATEGY,
            BestPracticesLayoutStrategy(),
            None,
            None,
            TEST_CATALOG_TEMPLATE,
            BEST_PRACTICE_CATALOG_TEMPLATE,
        ),
        (
            STRATEGY,
            None,
            BestPracticesLayoutStrategy(),
            None,
            BEST_PRACTICE_CATALOG_TEMPLATE,
            TEST_CATALOG_TEMPLATE,
        ),
        (
            STRATEGY,
            None,
            None,
            BestPracticesLayoutStrategy(),
            TEST_CATALOG_TEMPLATE,
            BEST_PRACTICE_CATALOG_TEMPLATE,
        ),
    ],
)
def test_add_child_layout_strategy(
    root_strategy: HrefLayoutStrategy,
    sub_strategy: HrefLayoutStrategy,
    provided_root_strategy: HrefLayoutStrategy,
    provided_sub_strategy: HrefLayoutStrategy,
    root_template: str,
    sub_template: str,
) -> None:
    """Test for layout strategy when adding a child.

    If no layout strategy is specified, children HREF should
    always follow BestPracticesLayoutStrategy.
    If only root strategy is set, all children HREFs should
    follow than strategy.
    If root and child strategy are set, root and child children
    may follow different strategies.
    Strategy provided to `add_child` always overrides other settings.
    """

    base_url = "http://example.com"
    catalog = Catalog(
        id="test",
        description="test desc",
        href=f"{base_url}/catalog.json",
        strategy=root_strategy,
    )
    subcat = Catalog(id="subcat", description="subcat desc", strategy=sub_strategy)
    subsubcat = Catalog(id="subsubcat", description="subsubcat desc")

    catalog.add_child(subcat, strategy=provided_root_strategy)
    subcat.add_child(subsubcat, strategy=provided_sub_strategy)

    root_template = root_template.format(
        id=subcat.id, description=subcat.description
    ).replace("$", "")
    sub_template = sub_template.format(
        id=subsubcat.id, description=subsubcat.description
    ).replace("$", "")

    assert subcat.self_href == f"{base_url}/{root_template}/catalog.json"
    assert (
        subsubcat.self_href == f"{base_url}/{root_template}/{sub_template}/catalog.json"
    )


@pytest.mark.parametrize(
    "root_strategy,sub_strategy,provided_root_strategy,"
    "root_template,sub_template,norm_template",
    [
        (
            None,
            None,
            None,
            BEST_PRACTICE_CATALOG_TEMPLATE,
            BEST_PRACTICE_CATALOG_TEMPLATE,
            BEST_PRACTICE_CATALOG_TEMPLATE,
        ),
        (
            STRATEGY,
            None,
            None,
            TEST_CATALOG_TEMPLATE,
            TEST_CATALOG_TEMPLATE,
            TEST_CATALOG_TEMPLATE,
        ),
        (
            STRATEGY,
            BestPracticesLayoutStrategy(),
            None,
            TEST_CATALOG_TEMPLATE,
            BEST_PRACTICE_CATALOG_TEMPLATE,
            TEST_CATALOG_TEMPLATE,
        ),
        (
            STRATEGY,
            None,
            BestPracticesLayoutStrategy(),
            TEST_CATALOG_TEMPLATE,
            TEST_CATALOG_TEMPLATE,
            BEST_PRACTICE_CATALOG_TEMPLATE,
        ),
    ],
)
def test_add_child_layout_strategy_normalize(
    root_strategy: HrefLayoutStrategy,
    sub_strategy: HrefLayoutStrategy,
    provided_root_strategy: HrefLayoutStrategy,
    root_template: str,
    sub_template: str,
    norm_template: str,
) -> None:
    """Test for layout strategy when adding a child and normalizing HREFs.

    If no layout strategy is specified, children HREF
    should always follow BestPracticesLayoutStrategy.
    If only root strategy is set, all children HREFs
    should follow than strategy.
    If root and child strategy are set, root strategy
    overrides child strategy.
    Strategy provided to `normalize_href` always
    overrides other settings.
    """

    base_url = "http://example.com"
    catalog = Catalog(
        id="test",
        description="test desc",
        href=f"{base_url}/catalog.json",
        strategy=root_strategy,
    )
    subcat = Catalog(id="subcat", description="subcat desc", strategy=sub_strategy)
    subsubcat = Catalog(id="subsubcat", description="subsubcat desc")
    catalog.add_child(subcat)
    subcat.add_child(subsubcat)

    _root_template = root_template.format(
        id=subcat.id, description=subcat.description
    ).replace("$", "")
    _sub_template = sub_template.format(
        id=subsubcat.id, description=subsubcat.description
    ).replace("$", "")

    assert subcat.self_href == f"{base_url}/{_root_template}/catalog.json"
    assert (
        subsubcat.self_href
        == f"{base_url}/{_root_template}/{_sub_template}/catalog.json"
    )

    catalog.normalize_hrefs(base_url, strategy=provided_root_strategy)

    _root_template = norm_template.format(
        id=subcat.id, description=subcat.description
    ).replace("$", "")
    _sub_template = norm_template.format(
        id=subsubcat.id, description=subsubcat.description
    ).replace("$", "")

    assert subcat.self_href == f"{base_url}/{_root_template}/catalog.json"
    assert (
        subsubcat.self_href
        == f"{base_url}/{_root_template}/{_sub_template}/catalog.json"
    )


@pytest.mark.parametrize(
    "root_strategy,provided_strategy,template",
    [
        (None, None, BEST_PRACTICE_ITEM_TEMPLATE),
        (STRATEGY, None, TEST_ITEM_TEMPLATE),
        (STRATEGY, BestPracticesLayoutStrategy(), BEST_PRACTICE_ITEM_TEMPLATE),
    ],
)
def test_add_item_layout_strategy(
    root_strategy: HrefLayoutStrategy,
    provided_strategy: HrefLayoutStrategy,
    template: str,
) -> None:
    """Test for layout strategy when adding an item.

    If no layout strategy is specified, item HREF
    should always follow BestPracticesLayoutStrategy.
    If only root strategy is set, all item HREFs
    should follow than strategy.
    Strategy provided to `add_item` always overrides other settings.
    """

    base_url = "http://example.com"
    item_id = "item_id"
    catalog = Catalog(
        id="test",
        description="test desc",
        href=f"{base_url}/catalog.json",
        strategy=root_strategy,
    )
    item = Item(
        id=item_id,
        geometry={
            "type": "Polygon",
            "coordinates": [
                [
                    [180.0, -90.0],
                    [180.0, 90.0],
                    [-180.0, 90.0],
                    [-180.0, -90.0],
                    [180.0, -90.0],
                ]
            ],
        },
        bbox=[-180, -90, 180, 90],
        datetime=datetime(2023, 1, 1),
        properties={},
        assets={
            "data": Asset(
                href="http://example.com/assets/data.tif",
                roles=["data"],
                title="DATA",
            )
        },
    )

    catalog.add_item(item, strategy=provided_strategy)

    template = template.format(id=item.id).replace("$", "")

    assert item.self_href == f"{base_url}/{template}/{item_id}.json"


def test_APILayoutStrategy_requires_root_to_be_url(
    catalog: Catalog, collection: Collection, item: Item
) -> None:
    collection.add_item(item)
    catalog.add_child(collection)
    with pytest.raises(
        pystac.errors.STACError,
        match="When using APILayoutStrategy the root_href must be a URL",
    ):
        catalog.normalize_hrefs(root_href="issues-1486", strategy=APILayoutStrategy())


def test_get_child_links_cares_about_media_type(catalog: pystac.Catalog) -> None:
    catalog.links.extend(
        [
            pystac.Link(
                rel="child", target="./child-1.json", media_type="application/json"
            ),
            pystac.Link(
                rel="child", target="./child-2.json", media_type="application/geo+json"
            ),
            pystac.Link(rel="child", target="./child-3.json"),
            # this one won't get counted since it's the wrong media_type
            pystac.Link(rel="child", target="./child.html", media_type="text/html"),
        ]
    )

    assert len(catalog.get_child_links()) == 3


def test_get_item_links_cares_about_media_type(catalog: pystac.Catalog) -> None:
    catalog.links.extend(
        [
            pystac.Link(
                rel="item", target="./item-1.json", media_type="application/json"
            ),
            pystac.Link(
                rel="item", target="./item-2.json", media_type="application/geo+json"
            ),
            pystac.Link(rel="item", target="./item-3.json"),
            # this one won't get counted since it's the wrong media_type
            pystac.Link(rel="item", target="./item.html", media_type="text/html"),
        ]
    )

    assert len(catalog.get_item_links()) == 3


def test_get_root_link_cares_about_media_type(catalog: pystac.Catalog) -> None:
    catalog.links.insert(
        0, pystac.Link(rel="root", target="./self.json", media_type="text/html")
    )
    root_link = catalog.get_root_link()
    assert root_link and root_link.target != "./self.json"


def test_clone_extra_fields(catalog: Catalog) -> None:
    catalog.extra_fields["foo"] = "bar"
    cloned = catalog.clone()
    assert cloned.extra_fields["foo"] == "bar"


def test_warn_if_next_link_present(catalog: Catalog) -> None:
    catalog.links.append(Link(rel="next", target="./next.json"))
    with pytest.warns(UserWarning):
        _ = list(catalog.get_children())



import pytest

import pystac
from pystac import ExtensionTypeError
from pystac.cache import CollectionCache
from pystac.extensions.item_assets import ItemAssetsExtension
from pystac.extensions.view import ViewExtension
from pystac.serialization import (
    identify_stac_object,
    identify_stac_object_type,
    merge_common_properties,
    migrate_to_latest,
)
from pystac.utils import get_required, str_to_datetime
from tests.utils import TestCases
from tests.utils.test_cases import ExampleInfo


class TestMigrate:
    @pytest.mark.parametrize("example", TestCases.get_examples_info())
    def test_migrate(self, example: ExampleInfo) -> None:
        collection_cache = CollectionCache()
        path = example.path

        d = pystac.StacIO.default().read_json(path)
        if identify_stac_object_type(d) == pystac.STACObjectType.ITEM:
            merge_common_properties(
                d, json_href=path, collection_cache=collection_cache
            )

        info = identify_stac_object(d)

        migrated_d = migrate_to_latest(d, info)

        migrated_info = identify_stac_object(migrated_d)

        assert migrated_info.object_type == info.object_type
        assert (
            migrated_info.version_range.latest_valid_version()
            == pystac.get_stac_version()
        )

        # Ensure all stac_extensions are schema URIs
        for e_id in migrated_d["stac_extensions"]:
            assert e_id.endswith(".json"), f"{e_id} is not a JSON schema URI"

    def test_migrates_removed_extension(self) -> None:
        item = pystac.Item.from_file(
            TestCases.get_path(
                "data-files/examples/0.8.1/extensions/sar/examples/sentinel1.json"
            )
        )
        assert "dtr" not in item.stac_extensions
        assert item.common_metadata.start_datetime == str_to_datetime(
            "2018-11-03T23:58:55.121559Z"
        )

    def test_migrates_added_extension(self) -> None:
        item = pystac.Item.from_file(
            TestCases.get_path(
                "data-files/examples/0.8.1/item-spec/examples/planet-sample.json"
            )
        )
        assert ViewExtension.has_extension(item)
        view_ext = ViewExtension.ext(item)
        assert view_ext.sun_azimuth, 101.8
        assert view_ext.sun_elevation, 58.8
        assert view_ext.off_nadir, 1

    def test_migrates_removes_extension(self) -> None:
        collection = pystac.Collection.from_file(
            TestCases.get_path(
                "data-files/examples/0.9.0/extensions/asset/"
                "examples/example-landsat8.json"
            )
        )

        assert ItemAssetsExtension.get_schema_uri() not in collection.stac_extensions
        assert not ItemAssetsExtension.has_extension(collection)
        assert "item_assets" in collection.extra_fields

        assert collection.stac_extensions == []
        assert collection.item_assets["thumbnail"].title == "Thumbnail"

    def test_migrates_pre_1_0_0_rc1_stats_summary(self) -> None:
        collection = pystac.Collection.from_file(
            TestCases.get_path(
                "data-files/examples/1.0.0-beta.2/collection-spec/"
                "examples/sentinel2.json"
            )
        )
        datetime_summary = get_required(
            collection.summaries.get_range("datetime"), collection.summaries, "datetime"
        )
        assert datetime_summary.minimum == "2015-06-23T00:00:00Z"
        assert datetime_summary.maximum, "2019-07-10T13:44:56Z"

    def test_should_raise_exception_when_passing_invalid_extension_object(
        self,
    ) -> None:
        with pytest.raises(
            ExtensionTypeError,
            match=r"^ItemAssetsExtension does not apply to type 'object'$",
        ):
            ItemAssetsExtension.ext(object())  # type: ignore


def test_migrate_works_even_if_stac_extensions_is_null(
    test_case_1_catalog: pystac.Catalog,
) -> None:
    collection = list(test_case_1_catalog.get_all_collections())[0]
    collection_dict = collection.to_dict()
    collection_dict["stac_extensions"] = None

    pystac.Collection.from_dict(collection_dict, migrate=True)


def test_migrate_updates_license_from_various() -> None:
    path = TestCases.get_path("data-files/examples/1.0.0/collectionless-item.json")

    item = pystac.Item.from_file(path)
    assert item.properties["license"] == "other"


def test_migrate_updates_license_from_proprietary() -> None:
    path = TestCases.get_path(
        "data-files/examples/1.0.0/collection-only/collection.json"
    )

    collection = pystac.Collection.from_file(path)
    assert collection.license == "other"



"""Implements the :stac-ext:`Render Extension <render>`."""

from __future__ import annotations

from typing import Any, Generic, Literal, TypeVar

import pystac
from pystac.extensions.base import ExtensionManagementMixin, PropertiesExtension
from pystac.extensions.hooks import ExtensionHooks
from pystac.utils import get_required, map_opt

#: Generalized version of :class:`~pystac.Collection` or :class:`~pystac.Item`
T = TypeVar("T", pystac.Collection, pystac.Item)

SCHEMA_URI_PATTERN: str = (
    "https://stac-extensions.github.io/render/v{version}/schema.json"
)
DEFAULT_VERSION: str = "2.0.0"

SUPPORTED_VERSIONS = [DEFAULT_VERSION]

RENDERS_PROP = "renders"


class Render:
    """Parameters for creating a rendered view of assets."""

    properties: dict[str, Any]

    def __init__(self, properties: dict[str, Any]) -> None:
        self.properties = properties

    @property
    def assets(self) -> list[str]:
        """
        List of asset keys referencing the assets that are
        used to make the rendering.
        """
        return get_required(self.properties.get("assets"), self, "assets")

    @assets.setter
    def assets(self, v: list[str]) -> None:
        self.properties["assets"] = v

    @property
    def title(self) -> str | None:
        """Title of the rendering"""
        return self.properties.get("title")

    @title.setter
    def title(self, v: str | None) -> None:
        if v is not None:
            self.properties["title"] = v
        else:
            self.properties.pop("title", None)

    @property
    def rescale(self) -> list[list[float]] | None:
        """A list of min/max value pairs to rescale each asset by, e.g.
        ``[[0, 5000], [0, 7000], [0, 9000]]``. If not provided, the
        assets will not be rescaled.
        """
        return self.properties.get("rescale")

    @rescale.setter
    def rescale(self, v: list[list[float]] | None) -> None:
        if v is not None:
            self.properties["rescale"] = v
        else:
            self.properties.pop("rescale", None)

    @property
    def nodata(self) -> float | str | None:
        """Nodata value."""
        return self.properties.get("nodata")

    @nodata.setter
    def nodata(self, v: float | str | None) -> None:
        if v is not None:
            self.properties["nodata"] = v
        else:
            self.properties.pop("nodata", None)

    @property
    def colormap_name(self) -> str | None:
        """Name of color map to apply to the render.
        See: https://matplotlib.org/stable/gallery/color/colormap_reference.html
        """
        return self.properties.get("colormap_name")

    @colormap_name.setter
    def colormap_name(self, v: str | None) -> None:
        if v is not None:
            self.properties["colormap_name"] = v
        else:
            self.properties.pop("colormap_name", None)

    @property
    def colormap(self) -> dict[str, Any] | None:
        """A dictionary containing a custom colormap definition.
        See: https://developmentseed.org/titiler/advanced/rendering/#custom-colormaps
        """
        return self.properties.get("colormap")

    @colormap.setter
    def colormap(self, v: dict[str, Any] | None) -> None:
        if v is not None:
            self.properties["colormap"] = v
        else:
            self.properties.pop("colormap", None)

    @property
    def color_formula(self) -> str | None:
        """A string containing a color formula to apply
        color corrections to images. Useful for reducing
        artefacts like atmospheric haze, dark shadows, or
        muted colors.
        See: https://developmentseed.org/titiler/advanced/rendering/#color-formula
        """
        return self.properties.get("color_formula")

    @color_formula.setter
    def color_formula(self, v: str | None) -> None:
        if v is not None:
            self.properties["color_formula"] = v
        else:
            self.properties.pop("color_formula", None)

    @property
    def resampling(self) -> str | None:
        """Resampling algorithm to apply to the referenced assets. See GDAL
        resampling algorithm for some examples.
        See: https://gdal.org/en/latest/programs/gdalwarp.html#cmdoption-gdalwarp-r
        """
        return self.properties.get("resampling")

    @resampling.setter
    def resampling(self, v: str | None) -> None:
        if v is not None:
            self.properties["resampling"] = v
        else:
            self.properties.pop("resampling", None)

    @property
    def expression(self) -> str | None:
        """Band arithmetic formula to apply to the referenced assets."""
        return self.properties.get("expression")

    @expression.setter
    def expression(self, v: str | None) -> None:
        if v is not None:
            self.properties["expression"] = v
        else:
            self.properties.pop("expression", None)

    @property
    def minmax_zoom(self) -> list[int] | None:
        """Zoom level range applicable for the visualization, e.g. ``[2, 18]``."""
        return self.properties.get("minmax_zoom")

    @minmax_zoom.setter
    def minmax_zoom(self, v: list[int] | None) -> None:
        if v is not None:
            self.properties["minmax_zoom"] = v
        else:
            self.properties.pop("minmax_zoom", None)

    def apply(
        self,
        assets: list[str],
        title: str | None = None,
        rescale: list[list[float]] | None = None,
        nodata: float | str | None = None,
        colormap_name: str | None = None,
        colormap: dict[str, Any] | None = None,
        color_formula: str | None = None,
        resampling: str | None = None,
        expression: str | None = None,
        minmax_zoom: list[int] | None = None,
    ) -> None:
        """Set the properties for a new Render.

        Args:
            assets : List of asset keys referencing the assets that are
                used to make the rendering.
            title : Title of the rendering.
            rescale : A list of min/max value pairs to rescale each asset by, e.g.
                ``[[0, 5000], [0, 7000], [0, 9000]]``. If not provided, the
                assets will not be rescaled.
            nodata : Nodata value.
            colormap_name : Name of color map to apply to the render.
                https://matplotlib.org/stable/gallery/color/colormap_reference.html
            colormap : A dictionary containing a custom colormap definition.
                https://developmentseed.org/titiler/advanced/rendering/#custom-colormaps
            color_formula : A string containing a color formula to apply
                color corrections to images. Useful for reducing
                artifacts like atmospheric haze, dark shadows, or
                muted colors.
                https://developmentseed.org/titiler/advanced/rendering/#color-formula
            resampling : Resampling algorithm to apply to the referenced assets. See
                GDAL resampling algorithm for some examples.
                https://gdal.org/en/latest/programs/gdalwarp.html#cmdoption-gdalwarp-r
            expression : Band arithmetic formula to apply to the referenced assets.
            minmax_zoom : Zoom level range applicable for the visualization,
                e.g. ``[2, 18]``.
        """
        self.assets = assets
        self.title = title
        self.rescale = rescale
        self.nodata = nodata
        self.colormap_name = colormap_name
        self.colormap = colormap
        self.color_formula = color_formula
        self.resampling = resampling
        self.expression = expression
        self.minmax_zoom = minmax_zoom

    @classmethod
    def create(
        cls,
        assets: list[str],
        title: str | None = None,
        rescale: list[list[float]] | None = None,
        nodata: float | str | None = None,
        colormap_name: str | None = None,
        colormap: dict[str, Any] | None = None,
        color_formula: str | None = None,
        resampling: str | None = None,
        expression: str | None = None,
        minmax_zoom: list[int] | None = None,
    ) -> Render:
        """Create a new Render.

        Args:
            assets: List of asset keys referencing the assets that are
                used to make the rendering.
            title: Title of the rendering.
            rescale: A list of min/max value pairs to rescale each asset by, e.g.
                ``[[0, 5000], [0, 7000], [0, 9000]]``. If not provided, the
                assets will not be rescaled.
            nodata: Nodata value.
            colormap_name: Name of color map to apply to the render.
                https://matplotlib.org/stable/gallery/color/colormap_reference.html
            colormap: A dictionary containing a custom colormap definition.
                https://developmentseed.org/titiler/advanced/rendering/#custom-colormaps
            color_formula: A string containing a color formula to apply
                color corrections to images. Useful for reducing
                artifacts like atmospheric haze, dark shadows, or
                muted colors.
                https://developmentseed.org/titiler/advanced/rendering/#color-formula
            resampling: Resampling algorithm to apply to the referenced assets. See
                GDAL resampling algorithm for some examples.
                https://gdal.org/en/latest/programs/gdalwarp.html#cmdoption-gdalwarp-r
            expression: Band arithmetic formula to apply to the referenced assets.
            minmax_zoom: Zoom level range applicable for the visualization,
                e.g. ``[2, 18]``.
        """
        c = cls({})
        c.apply(
            assets=assets,
            title=title,
            rescale=rescale,
            nodata=nodata,
            colormap_name=colormap_name,
            colormap=colormap,
            color_formula=color_formula,
            resampling=resampling,
            expression=expression,
            minmax_zoom=minmax_zoom,
        )
        return c

    def to_dict(self) -> dict[str, Any]:
        return self.properties

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, Render):
            raise NotImplementedError
        return self.properties == other.properties

    def __repr__(self) -> str:
        props = " ".join(
            [
                f"{key}={value}"
                for key, value in self.properties.items()
                if value is not None
            ]
        )
        return f"<Render {props}>"


class RenderExtension(
    Generic[T],
    PropertiesExtension,
    ExtensionManagementMixin[pystac.Item | pystac.Collection],
):
    """An abstract class that can be used to extend the properties of a
    :class:`~pystac.Collection` or :class:`~pystac.Item` with
    properties from the :stac-ext:`Render Extension <render>`. This class is
    generic over the type of STAC Object to be extended (e.g. :class:`~pystac.Item`,
    :class:`~pystac.Collection`).

    To create a concrete instance of :class:`RenderExtension`, use the
    :meth:`RenderExtension.ext` method. For example:

    .. code-block:: python

        >>> item: pystac.Item = ...
        >>> xr_ext = RenderExtension.ext(item)

    """

    name: Literal["render"] = "render"

    @classmethod
    def get_schema_uri(cls) -> str:
        return SCHEMA_URI_PATTERN.format(version=DEFAULT_VERSION)

    @classmethod
    def ext(cls, obj: T, add_if_missing: bool = False) -> RenderExtension[T]:
        """Extend the given STAC Object with properties from the
        :stac-ext:`Render Extension <render>`.

        This extension can be applied to instances of :class:`~pystac.Collection`
        or :class:`~pystac.Item`.

        Raises:
            pystac.ExtensionTypeError : If an invalid object type is passed.
        """
        if isinstance(obj, pystac.Collection):
            cls.ensure_has_extension(obj, add_if_missing)
            return CollectionRenderExtension(obj)
        elif isinstance(obj, pystac.Item):
            cls.ensure_has_extension(obj, add_if_missing)
            return ItemRenderExtension(obj)
        else:
            raise pystac.ExtensionTypeError(
                f"RenderExtension does not apply to type '{type(obj).__name__}'"
            )

    def apply(
        self,
        renders: dict[str, Render],
    ) -> None:
        """Applies the render extension fields to the extended
        object.

        Args:
            renders: a dictionary mapping render names to
                :class: `~pystac.extensions.render.Render` objects.
        """
        self.renders = renders

    @property
    def renders(self) -> dict[str, Render]:
        """A dictionary where each key is the name of a render and each
        value is a :class:`~Render` object.
        """
        renders: dict[str, dict[str, Any]] = get_required(
            self._get_property(RENDERS_PROP, dict[str, dict[str, Any]]),
            self,
            RENDERS_PROP,
        )
        return {k: Render(v) for k, v in renders.items()}

    @renders.setter
    def renders(self, v: dict[str, Render]) -> None:
        self._set_property(
            RENDERS_PROP,
            map_opt(lambda renders: {k: r.to_dict() for k, r in renders.items()}, v),
            pop_if_none=False,
        )


class CollectionRenderExtension(RenderExtension[pystac.Collection]):
    """A concrete implementation of :class:`RenderExtension` on a
    :class:`~pystac.Collection` that extends the properties of the Collection to include
    properties defined in the :stac-ext:`Render Extension <render>`.

    This class should generally not be instantiated directly. Instead, call
    :meth:`RenderExtension.ext` on an :class:`~pystac.Collection` to extend it.
    """

    def __init__(self, collection: pystac.Collection):
        self.collection = collection
        self.properties = collection.extra_fields

    def __repr__(self) -> str:
        return f"<CollectionRenderExtension Collection id={self.collection.id}>"


class ItemRenderExtension(RenderExtension[pystac.Item]):
    """A concrete implementation of :class:`RenderExtension` on a
    :class:`~pystac.Item` that extends the properties of the Item to include
    properties defined in the :stac-ext:`Render Extension <render>`.

    This class should generally not be instantiated directly. Instead, call
    :meth:`RenderExtension.ext` on an :class:`~pystac.Item` to extend it.
    """

    def __init__(self, item: pystac.Item):
        self.item = item
        self.properties = item.properties

    def __repr__(self) -> str:
        return f"<ItemRenderExtension Item id={self.item.id}>"


class RenderExtensionHooks(ExtensionHooks):
    schema_uri: str = SCHEMA_URI_PATTERN.format(version=DEFAULT_VERSION)
    stac_object_types = {pystac.STACObjectType.COLLECTION, pystac.STACObjectType.ITEM}



import json

import pytest

import pystac
from pystac import ExtensionTypeError, Item
from pystac.collection import Collection
from pystac.extensions.view import ViewExtension
from pystac.summaries import RangeSummary
from tests.utils import TestCases, assert_to_from_dict

EXAMPLE_URI = TestCases.get_path("data-files/view/example-landsat8.json")


@pytest.fixture
def item() -> Item:
    return pystac.Item.from_file(EXAMPLE_URI)


@pytest.fixture
def collection() -> Collection:
    return Collection.from_file(
        TestCases.get_path("data-files/view/collection-with-summaries.json")
    )


def test_to_from_dict() -> None:
    with open(EXAMPLE_URI) as f:
        d = json.load(f)
    assert_to_from_dict(pystac.Item, d)


def test_apply() -> None:
    item = next(TestCases.case_2().get_items(recursive=True))
    assert not ViewExtension.has_extension(item)

    ViewExtension.add_to(item)
    ViewExtension.ext(item).apply(
        off_nadir=1.0,
        incidence_angle=2.0,
        azimuth=3.0,
        sun_azimuth=4.0,
        sun_elevation=5.0,
    )

    assert ViewExtension.ext(item).off_nadir == 1.0
    assert ViewExtension.ext(item).incidence_angle == 2.0
    assert ViewExtension.ext(item).azimuth == 3.0
    assert ViewExtension.ext(item).sun_azimuth == 4.0
    assert ViewExtension.ext(item).sun_elevation == 5.0


@pytest.mark.vcr()
def test_validate_view(item: Item) -> None:
    assert ViewExtension.has_extension(item)
    item.validate()


@pytest.mark.vcr()
def test_off_nadir(item: Item) -> None:
    # Get
    assert "view:off_nadir" in item.properties
    view_off_nadir = ViewExtension.ext(item).off_nadir
    assert view_off_nadir is not None
    assert view_off_nadir == item.properties["view:off_nadir"]

    # Set
    ViewExtension.ext(item).off_nadir = view_off_nadir + 10
    assert view_off_nadir + 10 == item.properties["view:off_nadir"]

    # Get from Asset
    asset_no_prop = item.assets["blue"]
    asset_prop = item.assets["red"]
    assert (
        ViewExtension.ext(asset_no_prop).off_nadir == ViewExtension.ext(item).off_nadir
    )
    assert ViewExtension.ext(asset_prop).off_nadir == 3.0

    # Set to Asset
    asset_value = 13.0
    ViewExtension.ext(asset_no_prop).off_nadir = asset_value
    assert (
        ViewExtension.ext(asset_no_prop).off_nadir != ViewExtension.ext(item).off_nadir
    )
    assert ViewExtension.ext(asset_no_prop).off_nadir == asset_value

    # Validate
    item.validate()


@pytest.mark.vcr()
def test_incidence_angle(item: Item) -> None:
    # Get
    assert "view:incidence_angle" in item.properties
    view_incidence_angle = ViewExtension.ext(item).incidence_angle
    assert view_incidence_angle is not None
    assert view_incidence_angle == item.properties["view:incidence_angle"]
    # Set
    ViewExtension.ext(item).incidence_angle = view_incidence_angle + 10
    assert view_incidence_angle + 10 == item.properties["view:incidence_angle"]
    # Get from Asset
    asset_no_prop = item.assets["blue"]
    asset_prop = item.assets["red"]
    assert (
        ViewExtension.ext(asset_no_prop).incidence_angle
        == ViewExtension.ext(item).incidence_angle
    )
    assert ViewExtension.ext(asset_prop).incidence_angle == 4.0

    # Set to Asset
    asset_value = 14.0
    ViewExtension.ext(asset_no_prop).incidence_angle = asset_value
    assert (
        ViewExtension.ext(asset_no_prop).incidence_angle
        != ViewExtension.ext(item).incidence_angle
    )
    assert ViewExtension.ext(asset_no_prop).incidence_angle == asset_value

    # Validate
    item.validate()


@pytest.mark.vcr()
def test_azimuth(item: Item) -> None:
    # Get
    assert "view:azimuth" in item.properties
    view_azimuth = ViewExtension.ext(item).azimuth
    assert view_azimuth is not None
    assert view_azimuth == item.properties["view:azimuth"]

    # Set
    ViewExtension.ext(item).azimuth = view_azimuth + 100
    assert view_azimuth + 100 == item.properties["view:azimuth"]

    # Get from Asset
    asset_no_prop = item.assets["blue"]
    asset_prop = item.assets["red"]
    assert ViewExtension.ext(asset_no_prop).azimuth == ViewExtension.ext(item).azimuth
    assert ViewExtension.ext(asset_prop).azimuth == 5.0

    # Set to Asset
    asset_value = 15.0
    ViewExtension.ext(asset_no_prop).azimuth = asset_value
    assert ViewExtension.ext(asset_no_prop).azimuth != ViewExtension.ext(item).azimuth
    assert ViewExtension.ext(asset_no_prop).azimuth == asset_value

    # Validate
    item.validate()


@pytest.mark.vcr()
def test_sun_azimuth(item: Item) -> None:
    # Get
    assert "view:sun_azimuth" in item.properties
    view_sun_azimuth = ViewExtension.ext(item).sun_azimuth
    assert view_sun_azimuth is not None
    assert view_sun_azimuth == item.properties["view:sun_azimuth"]

    # Set
    ViewExtension.ext(item).sun_azimuth = view_sun_azimuth + 100
    assert view_sun_azimuth + 100 == item.properties["view:sun_azimuth"]
    # Get from Asset
    asset_no_prop = item.assets["blue"]
    asset_prop = item.assets["red"]
    assert (
        ViewExtension.ext(asset_no_prop).sun_azimuth
        == ViewExtension.ext(item).sun_azimuth
    )
    assert ViewExtension.ext(asset_prop).sun_azimuth == 1.0

    # Set to Asset
    asset_value = 11.0
    ViewExtension.ext(asset_no_prop).sun_azimuth = asset_value
    assert (
        ViewExtension.ext(asset_no_prop).sun_azimuth
        != ViewExtension.ext(item).sun_azimuth
    )
    assert ViewExtension.ext(asset_no_prop).sun_azimuth == asset_value

    # Validate
    item.validate()


@pytest.mark.vcr()
def test_sun_elevation(item: Item) -> None:
    # Get
    assert "view:sun_elevation" in item.properties
    view_sun_elevation = ViewExtension.ext(item).sun_elevation
    assert view_sun_elevation is not None
    assert view_sun_elevation == item.properties["view:sun_elevation"]

    # Set
    ViewExtension.ext(item).sun_elevation = view_sun_elevation + 10
    assert view_sun_elevation + 10 == item.properties["view:sun_elevation"]
    # Get from Asset
    asset_no_prop = item.assets["blue"]
    asset_prop = item.assets["red"]
    assert (
        ViewExtension.ext(asset_no_prop).sun_elevation
        == ViewExtension.ext(item).sun_elevation
    )
    assert ViewExtension.ext(asset_prop).sun_elevation == 2.0

    # Set to Asset
    asset_value = 12.0
    ViewExtension.ext(asset_no_prop).sun_elevation = asset_value
    assert (
        ViewExtension.ext(asset_no_prop).sun_elevation
        != ViewExtension.ext(item).sun_elevation
    )
    assert ViewExtension.ext(asset_no_prop).sun_elevation == asset_value

    # Validate
    item.validate()


def test_extension_not_implemented(item: Item) -> None:
    # Should raise exception if Item does not include extension URI
    item.stac_extensions.remove(ViewExtension.get_schema_uri())

    with pytest.raises(pystac.ExtensionNotImplemented):
        _ = ViewExtension.ext(item)

    # Should raise exception if owning Item does not include extension URI
    asset = item.assets["blue"]

    with pytest.raises(pystac.ExtensionNotImplemented):
        _ = ViewExtension.ext(asset)

    # Should succeed if Asset has no owner
    ownerless_asset = pystac.Asset.from_dict(asset.to_dict())
    _ = ViewExtension.ext(ownerless_asset)


def test_item_ext_add_to(item: Item) -> None:
    item.stac_extensions.remove(ViewExtension.get_schema_uri())
    assert ViewExtension.get_schema_uri() not in item.stac_extensions

    _ = ViewExtension.ext(item, add_if_missing=True)

    assert ViewExtension.get_schema_uri() in item.stac_extensions


def test_asset_ext_add_to(item: Item) -> None:
    item.stac_extensions.remove(ViewExtension.get_schema_uri())
    assert ViewExtension.get_schema_uri() not in item.stac_extensions
    asset = item.assets["blue"]

    _ = ViewExtension.ext(asset, add_if_missing=True)

    assert ViewExtension.get_schema_uri() in item.stac_extensions


def test_should_raise_exception_when_passing_invalid_extension_object() -> None:
    with pytest.raises(
        ExtensionTypeError, match=r"^ViewExtension does not apply to type 'object'$"
    ):
        # calling it wrong on purpose so ------v
        ViewExtension.ext(object())  # type: ignore


def test_get_off_nadir_summaries(collection: Collection) -> None:
    off_nadirs = ViewExtension.summaries(collection, True).off_nadir

    assert off_nadirs is not None

    assert {"minimum": 0.5, "maximum": 7.3} == off_nadirs.to_dict()


def test_get_incidence_angle_summaries(collection: Collection) -> None:
    incidence_angles = ViewExtension.summaries(collection, True).incidence_angle

    assert incidence_angles is not None

    assert {"minimum": 23, "maximum": 35} == incidence_angles.to_dict()


def test_get_azimuth_summaries(collection: Collection) -> None:
    azimuths = ViewExtension.summaries(collection, True).azimuth

    assert azimuths is not None

    assert {"minimum": 20, "maximum": 186} == azimuths.to_dict()


def test_get_sun_azimuth_summaries(collection: Collection) -> None:
    sun_azimuths = ViewExtension.summaries(collection, True).sun_azimuth

    assert sun_azimuths is not None

    assert {"minimum": 48, "maximum": 78} == sun_azimuths.to_dict()


def test_get_sun_elevation_summaries(collection: Collection) -> None:
    sun_elevations = ViewExtension.summaries(collection, True).sun_elevation

    assert sun_elevations is not None

    assert {"minimum": 10, "maximum": 45} == sun_elevations.to_dict()


def test_set_off_nadir_summaries(collection: Collection) -> None:
    view_summaries = ViewExtension.summaries(collection, True)

    view_summaries.off_nadir = RangeSummary(0, 10)
    assert {"minimum": 0, "maximum": 10} == view_summaries.off_nadir.to_dict()


def test_set_incidence_angle_summaries(collection: Collection) -> None:
    view_summaries = ViewExtension.summaries(collection, True)

    view_summaries.incidence_angle = RangeSummary(5, 15)
    assert {"minimum": 5, "maximum": 15} == view_summaries.incidence_angle.to_dict()


def test_set_azimuth_summaries(collection: Collection) -> None:
    view_summaries = ViewExtension.summaries(collection, True)

    view_summaries.azimuth = None
    assert view_summaries.azimuth is None


def test_set_sun_azimuth_summaries(collection: Collection) -> None:
    view_summaries = ViewExtension.summaries(collection, True)

    view_summaries.sun_azimuth = RangeSummary(210, 275)
    assert {"minimum": 210, "maximum": 275} == view_summaries.sun_azimuth.to_dict()


def test_set_sun_elevation_summaries(collection: Collection) -> None:
    view_summaries = ViewExtension.summaries(collection, True)

    view_summaries.sun_elevation = RangeSummary(-10, 38)
    assert {"minimum": -10, "maximum": 38} == view_summaries.sun_elevation.to_dict()


def test_summaries_adds_uri(collection: Collection) -> None:
    collection.stac_extensions = []
    with pytest.raises(
        pystac.ExtensionNotImplemented,
        match="Extension 'view' is not implemented",
    ):
        ViewExtension.summaries(collection, add_if_missing=False)

    _ = ViewExtension.summaries(collection, True)

    assert ViewExtension.get_schema_uri() in collection.stac_extensions

    ViewExtension.remove_from(collection)
    assert ViewExtension.get_schema_uri() not in collection.stac_extensions



from collections.abc import Iterable
from typing import Any, cast

import pystac
from pystac.cache import CollectionCache
from pystac.serialization.identify import STACVersionID
from pystac.utils import make_absolute_href


def merge_common_properties(
    item_dict: dict[str, Any],
    collection_cache: CollectionCache | None = None,
    json_href: str | None = None,
) -> bool:
    """Merges Collection properties into an Item.

    Note: This is only applicable to reading old STAC versions (pre 1.0.0-beta.1).

    Args:
        item_dict : JSON dict of the Item which properties should be merged
            into.
        collection_cache : Optional CollectionCache
            that will be used to read and write cached collections.
        json_href: The HREF of the file that this JSON comes from. Used
            to resolve relative paths.

    Returns:
        bool: True if Collection properties have been merged, otherwise False.
    """
    properties_merged = False

    collection: pystac.Collection | dict[str, Any] | None = None
    collection_href: str | None = None

    stac_version = item_dict.get("stac_version")

    # The commons extension was removed in 1.0.0-beta.1, so if this is an earlier STAC
    # item we don't have to bother with merging.
    if stac_version is not None and STACVersionID(stac_version) > "0.9.0":
        return False

    # Check to see if this is a 0.9.0 item that
    # doesn't extend the commons extension, in which case
    # we don't have to merge.
    if stac_version is not None and stac_version == "0.9.0":
        stac_extensions = item_dict.get("stac_extensions")
        if isinstance(stac_extensions, list):
            if "commons" not in stac_extensions:
                return False
        else:
            return False

    # Try the cache if we have a collection ID.
    collection_id = item_dict.get("collection")
    if collection_id is not None:
        if collection_cache is not None:
            collection = collection_cache.get_by_id(collection_id)

    # Next, try the collection link.
    if collection is None:
        # Account for 0.5 links, which were dicts
        if isinstance(item_dict["links"], dict):
            links = list(cast(Iterable[dict[str, Any]], item_dict["links"].values()))
        else:
            links = cast(list[dict[str, Any]], item_dict["links"])

        collection_link = next(
            (link for link in links if link["rel"] == pystac.RelType.COLLECTION), None
        )
        if collection_link is not None:
            collection_href = collection_link.get("href")
            if collection_href is not None:
                if json_href is not None:
                    collection_href = make_absolute_href(collection_href, json_href)
                if collection_cache is not None:
                    collection = collection_cache.get_by_href(collection_href)

                if collection is None:
                    collection = pystac.StacIO.default().read_json(collection_href)

    if collection is not None:
        collection_props: dict[str, Any] | None = None
        if isinstance(collection, pystac.Collection):
            collection_id = collection.id
            collection_props = collection.extra_fields.get("properties")
        elif isinstance(collection, dict):
            collection_id = collection["id"]
            if "properties" in collection:
                collection_props = collection["properties"]
        else:
            raise ValueError(
                "{} is expected to be a Collection or dict but is neither.".format(
                    collection
                )
            )

        if collection_props is not None:
            for k in collection_props:
                if k not in item_dict["properties"]:
                    properties_merged = True
                    item_dict["properties"][k] = collection_props[k]

        if (
            collection_cache is not None
            and collection_id is not None
            and not collection_cache.contains_id(collection_id)
        ):
            collection_cache.cache(collection, href=collection_href)

    return properties_merged



from pathlib import Path

import pytest

import pystac
from pystac import ExtensionTypeError, Item
from pystac.extensions.table import Column, TableExtension
from tests.utils import TestCases


@pytest.fixture
def table_item() -> Item:
    return pystac.Item.from_file(TestCases.get_path("data-files/table/item.json"))


@pytest.mark.vcr()
def test_validate(table_item: Item) -> None:
    table_item.validate()


def test_extension_not_implemented(table_item: Item) -> None:
    # Should raise exception if item does not include extension URI
    table_item.stac_extensions.remove(TableExtension.get_schema_uri())

    with pytest.raises(pystac.ExtensionNotImplemented):
        _ = TableExtension.ext(table_item)

    # Should raise exception if owning item does not include extension URI
    asset = table_item.assets["data"]

    with pytest.raises(pystac.ExtensionNotImplemented):
        _ = TableExtension.ext(asset)

    # Should succeed if Asset has no owner
    ownerless_asset = pystac.Asset.from_dict(asset.to_dict())
    _ = TableExtension.ext(ownerless_asset)


def test_item_ext_add_to(table_item: Item) -> None:
    table_item.stac_extensions.remove(TableExtension.get_schema_uri())

    _ = TableExtension.ext(table_item, add_if_missing=True)

    assert TableExtension.get_schema_uri() in table_item.stac_extensions


def test_asset_ext_add_to(table_item: Item) -> None:
    table_item.stac_extensions.remove(TableExtension.get_schema_uri())

    assert TableExtension.get_schema_uri() not in table_item.stac_extensions
    asset = table_item.assets["data"]

    _ = TableExtension.ext(asset, add_if_missing=True)
    assert TableExtension.get_schema_uri() in table_item.stac_extensions


def test_should_raise_when_passing_invalid_extension_object() -> None:
    with pytest.raises(
        ExtensionTypeError, match=r"^TableExtension does not apply to type 'object'$"
    ):
        # calling it wrong on purpose so ------v
        TableExtension.ext(object())  # type: ignore


def test_item_with_table_extension_is_serilalizable_and_roundtrips(
    tmp_path: Path,
    table_item: Item,
) -> None:
    # add column metadata
    tab_ext = TableExtension.ext(table_item, add_if_missing=True)
    columns = [
        Column({"name": "col_1", "type": "str"}),
        Column({"name": "col_2", "type": "byte_array"}),
    ]
    tab_ext.columns = columns
    table_item.save_object(dest_href=str(tmp_path / "item.json"))
    assert all(isinstance(c, Column) for c in tab_ext.columns)
    assert all(
        before.properties == after.properties
        for before, after in zip(columns, tab_ext.columns)
    )


@pytest.mark.parametrize(
    "schema_uri",
    (
        "https://stac-extensions.github.io/table/v1.0.0/schema.json",
        "https://stac-extensions.github.io/table/v1.0.1/schema.json",
        "https://stac-extensions.github.io/table/v1.1.0/schema.json",
    ),
)
def test_migrate(schema_uri: str, table_item: Item) -> None:
    item_dict = table_item.to_dict(include_self_link=False, transform_hrefs=False)
    item_dict["stac_extensions"] = [schema_uri]
    assert Item.from_dict(item_dict, migrate=True).stac_extensions == [
        "https://stac-extensions.github.io/table/v1.2.0/schema.json"
    ]



import logging
import sys


def setup_logging(level: int) -> None:
    for package in ["pystac", "tests"]:
        logger = logging.getLogger(package)
        logger.setLevel(level)

        formatter = logging.Formatter("[%(levelname)s] %(asctime)s - %(message)s")

        ch = logging.StreamHandler(sys.stdout)
        ch.setLevel(level)
        ch.setFormatter(formatter)
        logger.addHandler(ch)


setup_logging(logging.INFO)
